import functools
import io
import logging
import typing
from abc import ABC, abstractmethod
from typing import Union, Tuple, Callable, Optional, List, Sequence

import numpy as np
import pandas as pd
import torch
from torch import nn
from torch.nn import functional as F

from .torch_data import TensorScaler, VectorDataUtil, ClassificationVectorDataUtil, TorchDataSet, \
    TorchDataSetProvider, Tensoriser, TorchDataSetFromDataFrames, RuleBasedTensoriser, \
    TorchDataSetProviderFromVectorDataUtil
from .torch_enums import ClassificationOutputMode
from .torch_opt import NNOptimiser, NNLossEvaluatorRegression, NNLossEvaluatorClassification, NNOptimiserParams, TrainingInfo
from ..data import DataFrameSplitter
from ..normalisation import NormalisationMode
from ..util.dtype import to_float_array
from ..util.pickle import setstate
from ..util.string import ToStringMixin
from ..vector_model import VectorRegressionModel, VectorClassificationModel, TrainingContext

log: logging.Logger = logging.getLogger(__name__)


class MCDropoutCapableNNModule(nn.Module, ABC):
    """
    Base class for NN modules that are to support MC-Dropout.
    Support can be added by applying the _dropout function in the module's forward method.
    Then, to apply inference that samples results, call inferMCDropout rather than just using __call__.
    """

    def __init__(self) -> None:
        super().__init__()
        self._applyMCDropout = False
        self._pMCDropoutOverride = None

    def __setstate__(self, d: dict) -> None:
        if "_applyMCDropout" not in d:
            d["_applyMCDropout"] = False
        if "_pMCDropoutOverride" not in d:
            d["_pMCDropoutOverride"] = None
        super().__setstate__(d)

    def _dropout(self, x: torch.Tensor, p_training=None, p_inference=None) -> torch.Tensor:
        """
        This method is to to applied within the module's forward method to apply dropouts during training and/or inference.

        :param x: the model input tensor
        :param p_training: the probability with which to apply dropouts during training; if None, apply no dropout
        :param p_inference:  the probability with which to apply dropouts during MC-Dropout-based inference (via inferMCDropout,
            which may override the probability via its optional argument);
            if None, a dropout is not to be applied
        :return: a potentially modified version of x with some elements dropped out, depending on application context and dropout
            probabilities
        """
        if self.training and p_training is not None:
            return F.dropout(x, p_training)
        elif not self.training and self._applyMCDropout and p_inference is not None:
            return F.dropout(x, p_inference if self._pMCDropoutOverride is None else self._pMCDropoutOverride)
        else:
            return x

    def _enable_mc_dropout(self, enabled=True, p_mc_dropout_override=None) -> None:
        self._applyMCDropout = enabled
        self._pMCDropoutOverride = p_mc_dropout_override

    def infer_mc_dropout(self, x: Union[torch.Tensor, Sequence[torch.Tensor]], num_samples, p=None) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Applies inference using MC-Dropout, drawing the given number of samples.

        :param x: the model input (a tensor or tuple/list of tensors)
        :param num_samples: the number of samples to draw with MC-Dropout
        :param p: the dropout probability to apply, overriding the probability specified by the model's forward method; if None, use model's
            default
        :return: a pair (y, sd) where y the mean output tensor and sd is a tensor of the same dimension containing standard deviations
        """
        if type(x) not in (tuple, list):
            x = [x]
        results = []
        self._enable_mc_dropout(True, p_mc_dropout_override=p)
        try:
            for i in range(num_samples):
                y = self(*x)
                results.append(y)
        finally:
            self._enable_mc_dropout(False)
        results = torch.stack(results)
        mean = torch.mean(results, 0)
        stddev = torch.std(results, 0, unbiased=False)
        return mean, stddev


class TorchModel(ABC, ToStringMixin):
    """
    sensAI abstraction for torch models, which supports one-line training, allows for convenient model application,
    has basic mechanisms for data scaling, and soundly handles persistence (via pickle).
    An instance wraps a torch.nn.Module, which is constructed on demand during training via the factory method
    createTorchModule.
    """
    log: logging.Logger = log.getChild(__qualname__)

    def __init__(self, cuda=True) -> None:
        self.cuda: bool = cuda
        self.module: Optional[torch.nn.Module] = None
        self.outputScaler: Optional[TensorScaler] = None
        self.inputScaler: Optional[TensorScaler] = None
        self.trainingInfo: Optional[TrainingInfo] = None
        self._gpu: Optional[int] = None
        self._normalisationCheckThreshold: Optional[int] = 5

    def _tostring_exclude_private(self) -> bool:
        return True

    def set_torch_module(self, module: torch.nn.Module) -> None:
        self.module = module

    def set_normalisation_check_threshold(self, threshold: Optional[float]):
        self._normalisationCheckThreshold = threshold

    def get_module_bytes(self) -> bytes:
        bytes_io = io.BytesIO()
        torch.save(self.module, bytes_io)
        return bytes_io.getvalue()

    def set_module_bytes(self, model_bytes: bytes) -> None:
        model_file = io.BytesIO(model_bytes)
        self._load_model(model_file)

    def get_torch_module(self) -> torch.nn.Module:
        return self.module

    def _set_cuda_enabled(self, is_cuda_enabled: bool) -> None:
        self.cuda = is_cuda_enabled

    def _is_cuda_enabled(self) -> bool:
        return self.cuda

    def _load_model(self, model_file) -> None:  # TODO: complete type hints: what types are allowed for modelFile?
        try:
            self.module = torch.load(model_file)
            self._gpu = self._get_gpu_from_model_parameter_device()
        except:
            if self._is_cuda_enabled():
                if torch.cuda.device_count() > 0:
                    new_device = "cuda:0"
                else:
                    new_device = "cpu"
                self.log.warning(f"Loading of CUDA model failed, trying to map model to device {new_device}...")
                if type(model_file) != str:
                    model_file.seek(0)
                try:
                    self.module = torch.load(model_file, map_location=new_device)
                except:
                    self.log.warning(f"Failure to map model to device {new_device}, trying CPU...")
                    if new_device != "cpu":
                        new_device = "cpu"
                        self.module = torch.load(model_file, map_location=new_device)
                if new_device == "cpu":
                    self._set_cuda_enabled(False)
                    self._gpu = None
                else:
                    self._gpu = 0
                self.log.info(f"Model successfully loaded to {new_device}")
            else:
                raise

    @abstractmethod
    def create_torch_module(self) -> torch.nn.Module:
        pass

    def __getstate__(self) -> dict:
        state = dict(self.__dict__)
        del state["module"]
        state["modelBytes"] = self.get_module_bytes()
        return state

    def __setstate__(self, d: dict) -> None:
        # backward compatibility
        if "bestEpoch" in d:
            d["trainingInfo"] = TrainingInfo(best_epoch=d["bestEpoch"])
            del d["bestEpoch"]
        new_default_properties = {"_normalisationCheckThreshold": 5}

        model_bytes = None
        if "modelBytes" in d:
            model_bytes = d["modelBytes"]
            del d["modelBytes"]
        setstate(TorchModel, self, d, new_default_properties=new_default_properties)
        if model_bytes is not None:
            self.set_module_bytes(model_bytes)

    def apply(self,
            x: Union[torch.Tensor, np.ndarray, TorchDataSet, Sequence[torch.Tensor]],
            as_numpy: bool = True, create_batch: bool = False,
            mc_dropout_samples: Optional[int] = None,
            mc_dropout_probability: Optional[float] = None,
            scale_output: bool = False,
            scale_input: bool = False) -> Union[torch.Tensor, np.ndarray, Tuple]:
        """
        Applies the model to the given input tensor and returns the result

        :param x: the input tensor (either a batch or, if createBatch=True, a single data point), a data set or a tuple/list of tensors
            (if the model accepts more than one input).
            If it is a data set, it will be processed at once, so the data set must not be too large to be processed at once.
        :param as_numpy: flag indicating whether to convert the result to a numpy.array (if False, return tensor)
        :param create_batch: whether to add an additional tensor dimension for a batch containing just one data point
        :param mc_dropout_samples: if not None, apply MC-Dropout-based inference with the respective number of samples; if None, apply
            regular inference
        :param mc_dropout_probability: the probability with which to apply dropouts in MC-Dropout-based inference; if None, use model's
            default
        :param scale_output: whether to scale the output that is produced by the underlying model (using this instance's output scaler,
            if any)
        :param scale_input: whether to scale the input (using this instance's input scaler, if any) before applying the underlying model

        :return: an output tensor or, if MC-Dropout is applied, a pair (y, sd) where y the mean output tensor and sd is a tensor of the
            same dimension containing standard deviations
        """
        def extract(z):
            if scale_output:
                z = self.scaled_output(z)
            if self._is_cuda_enabled():
                z = z.cpu()
            z = z.detach()
            if as_numpy:
                z = z.numpy()
            return z

        model = self.get_torch_module()
        model.eval()

        if isinstance(x, TorchDataSet):
            x = next(x.iter_batches(x.size(), input_only=True, shuffle=False))
        elif isinstance(x, np.ndarray):
            x = to_float_array(x)
            x = torch.from_numpy(x).float()

        if type(x) not in (list, tuple):
            inputs = [x]
        else:
            inputs = x

        if self._is_cuda_enabled():
            torch.cuda.set_device(self._gpu)
            inputs = [t.cuda() for t in inputs]
        if scale_input:
            inputs = [self.inputScaler.normalise(t) for t in inputs]
        if create_batch:
            inputs = [t.view(1, *x.size()) for t in inputs]

        # check input normalisation
        if self._normalisationCheckThreshold is not None:
            for i, t in enumerate(inputs):
                if t.is_floating_point() and t.numel() > 0:  # skip any integer tensors (which typically contain lengths) and empty tensors
                    max_value = t.abs().max().item()
                    if max_value > self._normalisationCheckThreshold:
                        log.warning(f"Received value in input tensor {i} which is likely to not be correctly normalised: "
                                    f"maximum abs. value in tensor is %f" % max_value)
        if mc_dropout_samples is None:
            y = model(*inputs)
            return extract(y)
        else:
            y, stddev = model.inferMCDropout(x, mc_dropout_samples, p=mc_dropout_probability)
            return extract(y), extract(stddev)

    def apply_scaled(self, x: Union[torch.Tensor, np.ndarray, TorchDataSet, Sequence[torch.Tensor]],
            as_numpy: bool = True,
            create_batch: bool = False,
            mc_dropout_samples: Optional[int] = None,
            mc_dropout_probability: Optional[float] = None) \
            -> Union[torch.Tensor, np.ndarray]:
        """
        applies the model to the given input tensor and returns the scaled result (i.e. in the original scale)

        :param x: the input tensor(s) or data set
        :param as_numpy: flag indicating whether to convert the result to a numpy.array (if False, return tensor)
        :param create_batch: whether to add an additional tensor dimension for a batch containing just one data point
        :param mc_dropout_samples: if not None, apply MC-Dropout-based inference with the respective number of samples; if None, apply
            regular inference
        :param mc_dropout_probability: the probability with which to apply dropouts in MC-Dropout-based inference; if None, use model's
            default

        :return: a scaled output tensor or, if MC-Dropout is applied, a pair (y, sd) of scaled tensors, where
            y the mean output tensor and sd is a tensor of the same dimension containing standard deviations
        """
        return self.apply(x, scale_output=True, scale_input=True, as_numpy=as_numpy, create_batch=create_batch,
            mc_dropout_samples=mc_dropout_samples, mc_dropout_probability=mc_dropout_probability)

    def scaled_output(self, output: torch.Tensor) -> torch.Tensor:
        return self.outputScaler.denormalise(output)

    def _extract_params_from_data(self, data: TorchDataSetProvider) -> None:
        self.outputScaler = data.get_output_tensor_scaler()
        self.inputScaler = data.get_input_tensor_scaler()

    def fit(self, data: TorchDataSetProvider, nn_optimiser_params: NNOptimiserParams, strategy: "TorchModelFittingStrategy" = None) \
            -> None:
        """
        Fits this model using the given model and strategy

        :param data: a provider for the data with which to fit the model
        :param strategy: the fitting strategy; if None, use TorchModelFittingStrategyDefault.
            Pass your own strategy to perform custom fitting processes, e.g. process which involve multi-stage learning
        :param nn_optimiser_params: the parameters with which to create an optimiser which can be applied in the fitting strategy
        """
        self._extract_params_from_data(data)
        optimiser = NNOptimiser(nn_optimiser_params)
        if strategy is None:
            strategy = TorchModelFittingStrategyDefault()
        self.trainingInfo = strategy.fit(self, data, optimiser)
        self._gpu = self._get_gpu_from_model_parameter_device()

    def _get_gpu_from_model_parameter_device(self) -> Optional[int]:
        try:
            return next(self.module.parameters()).get_device()
        except:
            return None

    @property
    def best_epoch(self) -> Optional[int]:
        return self.trainingInfo.best_epoch if self.trainingInfo is not None else None

    @property
    def total_epochs(self) -> Optional[int]:
        return self.trainingInfo.total_epochs if self.trainingInfo is not None else None

    def _tostring_excludes(self) -> List[str]:
        return ['_gpu', 'module', 'trainingInfo', "inputScaler", "outputScaler"]

    def _tostring_additional_entries(self):
        return dict(bestEpoch=self.best_epoch, totalEpochs=self.total_epochs)


class TorchModelFittingStrategy(ABC):
    """
    Defines the interface for fitting strategies that can be used in TorchModel.fit
    """
    @abstractmethod
    def fit(self, model: TorchModel, data: TorchDataSetProvider, nn_optimiser: NNOptimiser) -> Optional[TrainingInfo]:
        pass


class TorchModelFittingStrategyDefault(TorchModelFittingStrategy):
    """
    Represents the default fitting strategy, which simply applies the given optimiser to the model and data
    """
    def fit(self, model: TorchModel, data: TorchDataSetProvider, nn_optimiser: NNOptimiser) -> Optional[TrainingInfo]:
        return nn_optimiser.fit(model, data)


class TorchModelFromModuleFactory(TorchModel):
    def __init__(self, module_factory: Callable[..., torch.nn.Module], *args, cuda: bool = True, **kwargs) -> None:
        super().__init__(cuda)
        self.args = args
        self.kwargs = kwargs
        self.moduleFactory = module_factory

    def create_torch_module(self) -> torch.nn.Module:
        return self.moduleFactory(*self.args, **self.kwargs)


class TorchModelFromModule(TorchModel):
    def __init__(self, module: torch.nn.Module, cuda: bool = True):
        super().__init__(cuda=cuda)
        self.module = module

    def create_torch_module(self) -> torch.nn.Module:
        return self.module


class TorchModelFactoryFromModule:
    """Represents a factory for the creation of a TorchModel based on a torch module"""
    def __init__(self, module: torch.nn.Module, cuda: bool = True):
        self.module = module
        self.cuda = cuda

    def __call__(self) -> TorchModel:
        return TorchModelFromModule(self.module, self.cuda)


class VectorTorchModel(TorchModel, ABC):
    """
    Base class for TorchModels that can be used within VectorModels, where the input and output dimensions
    are determined by the data
    """
    def __init__(self, cuda: bool = True) -> None:
        super().__init__(cuda=cuda)
        self.inputDim = None
        self.outputDim = None

    def _extract_params_from_data(self, data: TorchDataSetProvider) -> None:
        super()._extract_params_from_data(data)
        self.inputDim = data.get_input_dim()
        self.outputDim = data.get_model_output_dim()

    def create_torch_module(self) -> torch.nn.Module:
        return self.create_torch_module_for_dims(self.inputDim, self.outputDim)

    @abstractmethod
    def create_torch_module_for_dims(self, input_dim: int, output_dim: int) -> torch.nn.Module:
        """
        :param input_dim: the number of input dimensions as reported by the data set provider (number of columns
            in input data frame for default providers)
        :param output_dim: the number of output dimensions as reported by the data set provider (for default providers,
            this will be the number of columns in the output data frame or, for classification, the number of classes)
        :return: the torch module
        """
        pass


class TorchAutoregressiveResultHandler(ABC):
    """
    Supports the saving of predictions results such that subsequent predictions
    can build on earlier predictions, thus supporting autoregressive models.
    """

    @abstractmethod
    def clear_results(self):
        pass

    @abstractmethod
    def save_results(self, input_df: pd.DataFrame, results: np.ndarray) -> None:
        """
        Saves the regression results such that they can be used as input for subsequent prediction steps.
        The input will typically be processed by a feature generator or vectoriser, so the result
        should be stored in a place from which the respective feature generator or vectoriser can retrieve it.

        :param input_df: the input data frame for which results were obtained (number of rows corresponds to
            length of `results`)
        :param results: the results array, which is typically a 2D array where `results[i]` is an array
            containing the results for the i-th input row
        """
        pass


TTorchVectorRegressionModel = typing.TypeVar("TTorchVectorRegressionModel", bound="TorchVectorRegressionModel")


class TorchVectorRegressionModel(VectorRegressionModel):
    """
    Base class for the implementation of VectorRegressionModels based on TorchModels.
    An instance of this class will have an instance of TorchModel as the underlying model.
    """

    def __init__(self, torch_model_factory: Callable[[], TorchModel],
            normalisation_mode: NormalisationMode = NormalisationMode.NONE,
            nn_optimiser_params: Union[dict, NNOptimiserParams, None] = None) -> None:
        """
        :param torch_model_factory: the factory function with which to create the contained TorchModel instance that the instance is to
            encapsulate. For the instance to be picklable, this cannot be a lambda or locally defined function.
        :param normalisation_mode: the normalisation mode to apply to input data frames
        :param nn_optimiser_params: the parameters to apply in NNOptimiser during training
        """
        super().__init__()

        if nn_optimiser_params is None:
            nn_optimiser_params_instance = NNOptimiserParams()
        else:
            nn_optimiser_params_instance = NNOptimiserParams.from_dict_or_instance(nn_optimiser_params)
        if nn_optimiser_params_instance.loss_evaluator is None:
            nn_optimiser_params_instance.loss_evaluator = NNLossEvaluatorRegression(NNLossEvaluatorRegression.LossFunction.MSELOSS)

        self.torch_model_factory = torch_model_factory
        self.normalisationMode = normalisation_mode
        self.nnOptimiserParams = nn_optimiser_params_instance
        self.model: Optional[TorchModel] = None
        self.inputTensoriser: Optional[Tensoriser] = None
        self.outputTensoriser: Optional[Tensoriser] = None
        self.outputTensorToArrayConverter: Optional[OutputTensorToArrayConverter] = None
        self.torchDataSetProviderFactory: Optional[TorchDataSetProviderFactory] = None
        self.dataFrameSplitter: Optional[DataFrameSplitter] = None
        self._normalisationCheckThreshold = 5
        self.inferenceBatchSize: Optional[int] = None
        self.autoregressiveResultHandler: Optional[TorchAutoregressiveResultHandler] = None

    def __setstate__(self, state) -> None:
        if "modelClass" in state:  # old-style factory
            state["torch_model_factory"] = functools.partial(state["modelClass"], *state["modelArgs"], **state["modelKwArgs"])
            for k in ("modelClass", "modelArgs", "modelKwArgs"):
                del state[k]
        state["nnOptimiserParams"] = NNOptimiserParams.from_dict_or_instance(state["nnOptimiserParams"])
        new_optional_members = ["inputTensoriser", "torchDataSetProviderFactory", "dataFrameSplitter", "outputTensoriser",
            "outputTensorToArrayConverter", "inferenceBatchSize", "autoRegressiveResultHandler"]
        new_default_properties = {"_normalisationCheckThreshold": 5}
        setstate(TorchVectorRegressionModel, self, state, new_optional_properties=new_optional_members,
            new_default_properties=new_default_properties)

    @classmethod
    def from_module(cls, module: torch.nn.Module, cuda=True, normalisation_mode: NormalisationMode = NormalisationMode.NONE,
            nn_optimiser_params: Optional[NNOptimiserParams] = None) -> "TorchVectorRegressionModel":
        return cls(TorchModelFactoryFromModule(module=module, cuda=cuda), normalisation_mode=normalisation_mode,
            nn_optimiser_params=nn_optimiser_params)

    def _tostring_excludes(self) -> List[str]:
        excludes = super()._tostring_excludes()
        if self.model is not None:
            return excludes + ["modelClass", "modelArgs", "modelKwArgs"]
        else:
            return excludes

    def with_input_tensoriser(self: TTorchVectorRegressionModel, tensoriser: Tensoriser) -> TTorchVectorRegressionModel:
        """
        :param tensoriser: tensoriser to use in order to convert input data frames to (one or more) tensors.
            The default tensoriser directly converts the data frame's values (which is assumed to contain only scalars that
            can be coerced to floats) to a float tensor.
            The use of a custom tensoriser is necessary if a non-trivial conversion is necessary or if the data frame
            is to be converted to more than one input tensor.
        :return: self
        """
        self.inputTensoriser = tensoriser
        return self

    def with_output_tensoriser(self: TTorchVectorRegressionModel, tensoriser: RuleBasedTensoriser) -> TTorchVectorRegressionModel:
        """
        :param tensoriser: tensoriser to use in order to convert the output data frame to a tensor.
            The default output tensoriser directly converts the data frame's values to a float tensor.

            NOTE: It is required to be a rule-based tensoriser, because mechanisms that require fitting on the data
            and thus perform a data-dependendent conversion are likely to cause problems because they would need
            to be reversed at inference time (since the model will be trained on the converted values). If you require
            a transformation, use a target transformer, which will be applied before the tensoriser.
        :return: self
        """
        self.outputTensoriser = tensoriser
        return self

    def with_output_tensor_to_array_converter(self: TTorchVectorRegressionModel,
            output_tensor_to_array_converter: "OutputTensorToArrayConverter") -> TTorchVectorRegressionModel:
        """
        Configures the use of a custom converter from tensors to numpy arrays, which is applied during inference.
        A custom converter can be required, for example, to handle variable-length outputs (where the output tensor
        will typically contain unwanted padding). Note that since the converter is for inference only, it may be
        required to use a custom loss evaluator during training if the use of a custom converter is necessary.

        :param output_tensor_to_array_converter: the converter
        :return: self
        """
        self.outputTensorToArrayConverter = output_tensor_to_array_converter
        return self

    def with_torch_data_set_provider_factory(self: TTorchVectorRegressionModel,
            torch_data_set_provider_factory: "TorchDataSetProviderFactory") -> TTorchVectorRegressionModel:
        """
        :param torch_data_set_provider_factory: the torch data set provider factory, which is used to instantiate the provider which
            will provide the training and validation data sets from the input data frame that is passed in for learning.
            By default, TorchDataSetProviderFactoryRegressionDefault is used.
        :return: self
        """
        self.torchDataSetProviderFactory = torch_data_set_provider_factory
        return self

    def with_data_frame_splitter(self: TTorchVectorRegressionModel, data_frame_splitter: DataFrameSplitter) -> TTorchVectorRegressionModel:
        """
        :param data_frame_splitter: the data frame splitter which is used to split the input/output data frames that are passed for
            learning into a data frame that is used for training and a data frame that is used for validation.
            The input data frame is the data frame that is passed as input to the splitter, and the returned indices
            are used to split both the input and output data frames in the same way.
        :return: self
        """
        self.dataFrameSplitter = data_frame_splitter
        return self

    def with_normalisation_check_threshold(self: TTorchVectorRegressionModel, threshold: Optional[float]) -> TTorchVectorRegressionModel:
        """
        Defines a threshold with which to check inputs that are passed to the underlying neural network.
        Whenever an (absolute) input value exceeds the threshold, a warning is triggered.

        :param threshold: the threshold
        :return: self
        """
        self._normalisationCheckThreshold = threshold
        return self

    def with_autoregressive_result_handler(self: TTorchVectorRegressionModel,
            result_handler: TorchAutoregressiveResultHandler,
            inference_batch_size=1) -> TTorchVectorRegressionModel:
        """
        Adds a result handler which can be used to store prediction results such that subsequent predictions
        can use the prediction result, supporting autoregressive models.
        The autoregressive predictions are assumed to be handled in a single call to method :meth:`predict`,
        and the results will be stored for the duration of the call.
        For autoregressive predictions that build on earlier predictions, we must typically restrict
        the batch size such that predictions from the earlier batch can be saved and correctly reused
        as input for the subsequent predictions. The models input preprocessors (such as feature generators
        or vectorisers) must make ensure that the results being stored by the result handler are appropriately
        used as input.

        :param result_handler: the result handler
        :param inference_batch_size: the batch size to use for predictions
        :return: self
        """
        self.autoregressiveResultHandler = result_handler
        self.inferenceBatchSize = inference_batch_size
        return self

    def _create_torch_model(self) -> TorchModel:
        torch_model = self.torch_model_factory()
        torch_model.set_normalisation_check_threshold(self._normalisationCheckThreshold)
        return torch_model

    def _create_data_set_provider(self, inputs: pd.DataFrame, outputs: pd.DataFrame) -> TorchDataSetProvider:
        factory = self.torchDataSetProviderFactory
        if factory is None:
            factory = TorchDataSetProviderFactoryRegressionDefault()
        return factory.create_data_set_provider(inputs, outputs, self, self._trainingContext, input_tensoriser=self.inputTensoriser,
            output_tensoriser=self.outputTensoriser, data_frame_splitter=self.dataFrameSplitter)

    def _fit(self, inputs: pd.DataFrame, outputs: pd.DataFrame, weights: Optional[pd.Series] = None) -> None:
        self._warn_sample_weights_unsupported(False, weights)
        if self.inputTensoriser is not None:
            log.info(f"Fitting {self.inputTensoriser} ...")
            self.inputTensoriser.fit(inputs, model=self)
        self.model = self._create_torch_model()
        data_set_provider = self._create_data_set_provider(inputs, outputs)
        self.model.fit(data_set_provider, self.nnOptimiserParams)

    def _predict_outputs_for_input_data_frame(self, inputs: pd.DataFrame) -> np.ndarray:
        tensorise_dynamically = False
        if self.autoregressiveResultHandler is not None:
            self.autoregressiveResultHandler.clear_results()
            tensorise_dynamically = True  # must be dynamically tensorised to allow inclusion of predicted results
        batch_size = self.nnOptimiserParams.batch_size if self.inferenceBatchSize is None else self.inferenceBatchSize
        results: List[np.ndarray] = []
        data_set = TorchDataSetFromDataFrames(inputs, None, self.model.cuda, input_tensoriser=self.inputTensoriser,
            tensorise_dynamically=tensorise_dynamically)
        start_idx = 0
        for input_batch in data_set.iter_batches(batch_size, input_only=True):
            if self.outputTensorToArrayConverter is None:
                result = self.model.apply_scaled(input_batch, as_numpy=True)
            else:
                output_batch = self.model.apply_scaled(input_batch, as_numpy=False)
                result = self.outputTensorToArrayConverter.convert(output_batch, input_batch)
            if self.autoregressiveResultHandler is not None:
                self.autoregressiveResultHandler.save_results(inputs.iloc[start_idx:start_idx+len(result)], result)
            start_idx += len(result)
            results.append(result)
        if self.autoregressiveResultHandler is not None:
            self.autoregressiveResultHandler.clear_results()
        return np.concatenate(results)

    def _predict(self, inputs: pd.DataFrame) -> pd.DataFrame:
        y_array = self._predict_outputs_for_input_data_frame(inputs)
        return pd.DataFrame(y_array, columns=self.get_model_output_variable_names())


TTorchVectorClassificationModel = typing.TypeVar("TTorchVectorClassificationModel", bound="TorchVectorClassificationModel")


class TorchVectorClassificationModel(VectorClassificationModel):
    """
    Base class for the implementation of VectorClassificationModels based on TorchModels.
    An instance of this class will have an instance of TorchModel as the underlying model.
    """
    def __init__(self, output_mode: ClassificationOutputMode,
            torch_model_factory: Callable[[], TorchModel],
            normalisation_mode: NormalisationMode = NormalisationMode.NONE,
            nn_optimiser_params: Optional[NNOptimiserParams] = None) -> None:
        """
        :param output_mode: specifies the nature of the output of the underlying neural network model
        :param torch_model_factory: the factory function with which to create the contained TorchModel instance that the instance is to
            encapsulate. For the instance to be picklable, this cannot be a lambda or locally defined function.
        :param normalisation_mode: the normalisation mode to apply to input data frames
        :param nn_optimiser_params: the parameters to apply in NNOptimiser during training
        """
        super().__init__()

        if nn_optimiser_params is None:
            nn_optimiser_params = NNOptimiserParams()
        if nn_optimiser_params.loss_evaluator is None:
            loss_function = NNLossEvaluatorClassification.LossFunction.default_for_output_mode(output_mode)
            nn_optimiser_params.loss_evaluator = NNLossEvaluatorClassification(loss_function)

        self.outputMode = output_mode
        self.torch_model_factory = torch_model_factory
        self.normalisationMode = normalisation_mode
        self.nnOptimiserParams: NNOptimiserParams = nn_optimiser_params
        self.model: Optional[TorchModel] = None
        self.inputTensoriser: Optional[Tensoriser] = None
        self.outputTensoriser: Optional[Tensoriser] = None
        self.torchDataSetProviderFactory: Optional[TorchDataSetProviderFactory] = None
        self.dataFrameSplitter: Optional[DataFrameSplitter] = None
        self._normalisationCheckThreshold = 5

    # noinspection DuplicatedCode
    def __setstate__(self, state) -> None:
        if "modelClass" in state:  # old-style factory
            state["torch_model_factory"] = functools.partial(state["modelClass"], *state["modelArgs"], **state["modelKwArgs"])
            for k in ("modelClass", "modelArgs", "modelKwArgs"):
                del state[k]
        state["nnOptimiserParams"] = NNOptimiserParams.from_dict_or_instance(state["nnOptimiserParams"])
        new_optional_members = ["inputTensoriser", "torchDataSetProviderFactory", "dataFrameSplitter", "outputTensoriser"]
        new_default_properties = {"outputMode": ClassificationOutputMode.PROBABILITIES, "_normalisationCheckThreshold": 5}
        setstate(TorchVectorClassificationModel, self, state, new_optional_properties=new_optional_members,
            new_default_properties=new_default_properties)

    @classmethod
    def from_module(cls, module: torch.nn.Module, output_mode: ClassificationOutputMode, cuda=True,
            normalisation_mode: NormalisationMode = NormalisationMode.NONE,
            nn_optimiser_params: Optional[NNOptimiserParams] = None) -> "TorchVectorClassificationModel":
        return cls(output_mode, TorchModelFactoryFromModule(module, cuda=cuda),
            normalisation_mode=normalisation_mode, nn_optimiser_params=nn_optimiser_params)

    def with_input_tensoriser(self: TTorchVectorClassificationModel, tensoriser: Tensoriser) -> TTorchVectorClassificationModel:
        """
        :param tensoriser: tensoriser to use in order to convert input data frames to (one or more) tensors.
            The default tensoriser directly converts the data frame's values (which is assumed to contain only scalars that
            can be coerced to floats) to a float tensor.
            The use of a custom tensoriser is necessary if a non-trivial conversion is necessary or if the data frame
            is to be converted to more than one input tensor.
        :return: self
        """
        self.inputTensoriser = tensoriser
        return self

    def with_output_tensoriser(self: TTorchVectorClassificationModel, tensoriser: RuleBasedTensoriser) -> TTorchVectorClassificationModel:
        """
        :param tensoriser: tensoriser to use in order to convert the output data frame to a tensor.
            NOTE: It is required to be a rule-based tensoriser, because mechanisms that require fitting on the data
            and thus perform a data-dependendent conversion are likely to cause problems because they would need
            to be reversed at inference time (since the model will be trained on the converted values). If you require
            a transformation, use a target transformer, which will be applied before the tensoriser.
        """
        self.outputTensoriser = tensoriser
        return self

    def with_torch_data_set_provider_factory(self: TTorchVectorClassificationModel,
            torch_data_set_provider_factory: "TorchDataSetProviderFactory") -> TTorchVectorClassificationModel:
        """
        :param torch_data_set_provider_factory: the torch data set provider factory, which is used to instantiate the provider which
            will provide the training and validation data sets from the input data frame that is passed in for learning.
            By default, TorchDataSetProviderFactoryClassificationDefault is used.
        :return: self
        """
        self.torchDataSetProviderFactory = torch_data_set_provider_factory
        return self

    def with_data_frame_splitter(self: TTorchVectorClassificationModel, data_frame_splitter: DataFrameSplitter) \
            -> TTorchVectorClassificationModel:
        """
        :param data_frame_splitter: the data frame splitter which is used to split the input/output data frames that are passed for
            learning into a data frame that is used for training and a data frame that is used for validation.
            The input data frame is the data frame that is passed as input to the splitter, and the returned indices
            are used to split both the input and output data frames in the same way.
        :return: self
        """
        self.dataFrameSplitter = data_frame_splitter
        return self

    def with_normalisation_check_threshold(self: TTorchVectorClassificationModel, threshold: Optional[float]) \
            -> TTorchVectorClassificationModel:
        """
        Defines a threshold with which to check inputs that are passed to the underlying neural network.
        Whenever an (absolute) input value exceeds the threshold, a warning is triggered.

        :param threshold: the threshold
        :return: self
        """
        self._normalisationCheckThreshold = threshold
        return self

    def _create_torch_model(self) -> TorchModel:
        torch_model = self.torch_model_factory()
        torch_model.set_normalisation_check_threshold(self._normalisationCheckThreshold)
        return torch_model

    def _create_data_set_provider(self, inputs: pd.DataFrame, outputs: pd.DataFrame) -> TorchDataSetProvider:
        factory = self.torchDataSetProviderFactory
        if factory is None:
            factory = TorchDataSetProviderFactoryClassificationDefault()
        return factory.create_data_set_provider(inputs, outputs, self, self._trainingContext, input_tensoriser=self.inputTensoriser,
            output_tensoriser=self.outputTensoriser, data_frame_splitter=self.dataFrameSplitter)

    def _fit_classifier(self, inputs: pd.DataFrame, outputs: pd.DataFrame, weights: Optional[pd.Series] = None) -> None:
        self._warn_sample_weights_unsupported(False, weights)
        if len(outputs.columns) != 1:
            raise ValueError("Expected one output dimension: the class labels")

        if self.inputTensoriser is not None:
            log.info(f"Fitting {self.inputTensoriser} ...")
            self.inputTensoriser.fit(inputs, model=self)

        # transform outputs: for each data point, the new output shall be the index in the list of labels
        labels: pd.Series = outputs.iloc[:, 0]
        outputs = pd.DataFrame([self._labels.index(l) for l in labels], columns=outputs.columns, index=outputs.index)

        self.model = self._create_torch_model()

        data_set_provider = self._create_data_set_provider(inputs, outputs)
        self.model.fit(data_set_provider, self.nnOptimiserParams)

    def _predict_outputs_for_input_data_frame(self, inputs: pd.DataFrame) -> torch.Tensor:
        batch_size = self.nnOptimiserParams.batch_size
        results = []
        data_set = TorchDataSetFromDataFrames(inputs, None, self.model.cuda, input_tensoriser=self.inputTensoriser)
        for inputBatch in data_set.iter_batches(batch_size, input_only=True):
            results.append(self.model.apply_scaled(inputBatch, as_numpy=False))
        return torch.cat(results, dim=0)

    def _predict_class_probabilities(self, inputs: pd.DataFrame) -> pd.DataFrame:
        y = self._predict_outputs_for_input_data_frame(inputs)
        if self.outputMode == ClassificationOutputMode.PROBABILITIES:
            pass
        elif self.outputMode == ClassificationOutputMode.LOG_PROBABILITIES:
            y = y.exp()
        elif self.outputMode == ClassificationOutputMode.UNNORMALISED_LOG_PROBABILITIES:
            y = y.softmax(dim=1)
        else:
            raise ValueError(f"Unhandled output mode {self.outputMode}")
        return pd.DataFrame(y.numpy(), columns=self._labels)

    def _tostring_excludes(self) -> List[str]:
        excludes = super()._tostring_excludes()
        if self.model is not None:
            return excludes + ["modelClass", "modelArgs", "modelKwArgs"]
        else:
            return excludes


class TorchDataSetProviderFactory(ABC):
    @abstractmethod
    def create_data_set_provider(self,
            inputs: pd.DataFrame,
            outputs: pd.DataFrame,
            model: Union[TorchVectorRegressionModel, TorchVectorClassificationModel],
            training_context: TrainingContext,
            input_tensoriser: Optional[Tensoriser],
            output_tensoriser: Optional[Tensoriser],
            data_frame_splitter: Optional[DataFrameSplitter]) -> TorchDataSetProvider:
        pass


class TorchDataSetProviderFactoryClassificationDefault(TorchDataSetProviderFactory):
    def __init__(self, tensorise_dynamically=False):
        """
        :param tensorise_dynamically: whether tensorisation shall take place on the fly whenever the provided data sets are iterated;
              if False, tensorisation takes place once in a precomputation stage (tensors must jointly fit into memory)
        """
        self.tensoriseDynamically = tensorise_dynamically

    def create_data_set_provider(self,
            inputs: pd.DataFrame,
            outputs: pd.DataFrame,
            model: TorchVectorClassificationModel,
            training_context: TrainingContext,
            input_tensoriser: Optional[Tensoriser],
            output_tensoriser: Optional[Tensoriser],
            data_frame_splitter: Optional[DataFrameSplitter]) -> TorchDataSetProvider:
        data_util = ClassificationVectorDataUtil(inputs, outputs, model.model.cuda, len(model._labels),  # TODO FIXME
            normalisation_mode=model.normalisationMode, input_tensoriser=input_tensoriser, output_tensoriser=output_tensoriser,
            data_frame_splitter=data_frame_splitter)
        return TorchDataSetProviderFromVectorDataUtil(data_util, model.model.cuda, tensorise_dynamically=self.tensoriseDynamically)


class TorchDataSetProviderFactoryRegressionDefault(TorchDataSetProviderFactory):
    def __init__(self, tensorise_dynamically=False):
        """
        :param tensorise_dynamically: whether tensorisation shall take place on the fly whenever the provided data sets are iterated;
              if False, tensorisation takes place once in a precomputation stage (tensors must jointly fit into memory)
        """
        self.tensoriseDynamically = tensorise_dynamically

    def create_data_set_provider(self, inputs: pd.DataFrame, outputs: pd.DataFrame, model: TorchVectorRegressionModel,
            training_context: TrainingContext, input_tensoriser: Optional[Tensoriser], output_tensoriser: Optional[Tensoriser],
            data_frame_splitter: Optional[DataFrameSplitter]) -> TorchDataSetProvider:
        data_util = VectorDataUtil(inputs, outputs, model.model.cuda, normalisation_mode=model.normalisationMode,
            input_tensoriser=input_tensoriser, output_tensoriser=output_tensoriser, data_frame_splitter=data_frame_splitter)
        return TorchDataSetProviderFromVectorDataUtil(data_util, model.model.cuda, tensorise_dynamically=self.tensoriseDynamically)


class OutputTensorToArrayConverter(ABC):
    @abstractmethod
    def convert(self, model_output: torch.Tensor, model_input: Union[torch.Tensor, Sequence[torch.Tensor]]) -> np.ndarray:
        """
        :param model_output: the output tensor generated by the model
        :param model_input: the input tensor(s) for which the model produced the output (which may provide relevant meta-data)
        :return: a numpy array of shape (N, D) where N=output.shape[0] is the number of data points and D is the number of
            variables predicted by the model 
        """
        pass



import functools
import re
from typing import Iterable, Dict
from glob import glob

from setuptools import setup, find_namespace_packages

# list of dependencies where ==/~= dependencies (used in requirements.txt and for the extras in requirements_*.txt) are relaxed:
# any later version is OK (as long as we are not aware of a concrete limitation - and once we are, we shall define
# the respective upper bound below)
DEPS_VERSION_LOWER_BOUND = [
    # main
    "pandas", "scipy", "numpy", "scikit-learn", "seaborn", "typing-extensions",
    # extra "torch"
    "torch", "torchtext",
    # extra "tensorflow"
    "tensorflow",
    # extra "lightgbm"
    "lightgbm",
    # extra "geoanalytics"
    "networkx", "Shapely", "geopandas", "utm",
    # extra "xgboost"
    "xgboost",
]
# upper bound: map dependency name to lowest exluded version
DEPS_VERSION_UPPER_BOUND_EXCLUSIVE: Dict[str, str] = {}


def relaxed_requirements(deps: Iterable[str]):
    """
    :param deps: the set of requirements
    :return: the set of updated requirements with the relaxations defined above applied
    """
    updated_deps = []
    for dep in deps:
        dep = dep.strip()
        if dep.startswith("#"):
            continue
        m = re.match(r'([\w-]+)[=~]=', dep)  # match package with == or ~= version spec
        if m:
            package = m.group(1)
            if package in DEPS_VERSION_LOWER_BOUND:
                dep = dep.replace("==", ">=").replace("~=", ">=")
            elif package in DEPS_VERSION_UPPER_BOUND_EXCLUSIVE:
                dep = dep.replace("==", ">=").replace("~=", ">=")
                dep += ",<" + DEPS_VERSION_UPPER_BOUND_EXCLUSIVE[package]
        updated_deps.append(dep)
    return updated_deps


def relaxed_requirements_from_file(path):
    with open(path, "r") as f:
        return relaxed_requirements(f.readlines())


# create extras requirements from requirements_*.txt, and add "full" extras which combines them all
extras_require = {}
for extras_requirements_file in glob("requirements_*.txt"):
    m = re.match(r"requirements_(\w+).txt", extras_requirements_file)
    extra_name = m.group(1)
    extras_require[extra_name] = relaxed_requirements_from_file(extras_requirements_file)
extras_require["full"] = functools.reduce(lambda x, y: x + y, list(extras_require.values()))


setup(
    name='sensai',
    package_dir={"": "src"},
    license="MIT",
    url="https://github.com/aai-institute/sensAI",
    packages=find_namespace_packages(where="src"),
    include_package_data=True,
    version='1.3.0',
    description='The Python library for sensible AI',
    install_requires=relaxed_requirements_from_file("requirements.txt"),
    dependency_links=["https://download.pytorch.org/whl/torch_stable.html"],
    setup_requires=["wheel"],
    extras_require=extras_require,
    author='appliedAI Institute gGmbh & jambit GmbH',
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License"
    ]
)





class Version:
    """
    Assists in checking the version of a Python package based on the __version__ attribute
    """
    def __init__(self, package):
        """
        :param package: the package object
        """
        self.components = package.__version__.split(".")

    def is_at_least(self, *components: int):
        """
        Checks this version against the given version components.
        This version object must contain at least the respective number of components

        :param components: version components in order (i.e. major, minor, patch, etc.)
        :return: True if the version is at least the given version, False otherwise
        """
        for i, desired_min_version in enumerate(components):
            actual_version = int(self.components[i])
            if actual_version < desired_min_version:
                return False
            elif actual_version > desired_min_version:
                return True
        return True

    def is_at_most(self, *components: int):
        """
        Checks this version against the given version components.
        This version object must contain at least the respective number of components

        :param components: version components in order (i.e. major, minor, patch, etc.)
        :return: True if the version is at most the given version, False otherwise
        """
        for i, desired_max_version in enumerate(components):
            actual_version = int(self.components[i])
            if actual_version > desired_max_version:
                return False
            elif actual_version < desired_max_version:
                return True
        return True

    def is_equal(self, *components: int):
        """
        Checks this version against the given version components.
        This version object must contain at least the respective number of components

        :param components: version components in order (i.e. major, minor, patch, etc.)
        :return: True if the version is the given version, False otherwise
        """
        return self.components[:len(components)] == list(components)



import collections
import logging
from typing import Sequence, Callable, TypeVar, Tuple, Optional, List, Any, Union, Dict

import matplotlib.figure
import matplotlib.ticker as plticker
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap

from sensai.util.pandas import SeriesInterpolation

log = logging.getLogger(__name__)

MATPLOTLIB_DEFAULT_FIGURE_SIZE = (6.4, 4.8)


class Color:
    def __init__(self, c: Any):
        """
        :param c: any color specification that is understood by matplotlib
        """
        self.rgba = matplotlib.colors.to_rgba(c)

    def darken(self, amount: float):
        """
        :param amount: amount to darken in [0,1], where 1 results in black and 0 leaves the color unchanged
        :return: the darkened color
        """
        import colorsys
        rgb = matplotlib.colors.to_rgb(self.rgba)
        h, l, s = colorsys.rgb_to_hls(*rgb)
        l *= amount
        rgb = colorsys.hls_to_rgb(h, l, s)
        return Color((*rgb, self.rgba[3]))

    def lighten(self, amount: float):
        """
        :param amount: amount to lighten in [0,1], where 1 results in white and 0 leaves the color unchanged
        :return: the lightened color
        """
        import colorsys
        rgb = matplotlib.colors.to_rgb(self.rgba)
        h, l, s = colorsys.rgb_to_hls(*rgb)
        l += (1-l) * amount
        rgb = colorsys.hls_to_rgb(h, l, s)
        return Color((*rgb, self.rgba[3]))

    def alpha(self, opacity: float) -> "Color":
        """
        Returns a new color with modified alpha channel (opacity)
        :param opacity: the opacity between 0 (transparent) and 1 (fully opaque)
        :return: the modified color
        """
        if not (0 <= opacity <= 1):
            raise ValueError(f"Opacity must be between 0 and 1, got {opacity}")
        return Color((*self.rgba[:3], opacity))

    def to_hex(self, keep_alpha=True) -> str:
        return matplotlib.colors.to_hex(self.rgba, keep_alpha)


class LinearColorMap:
    """
    Facilitates usage of linear segmented colour maps by combining a colour map (member `cmap`), which transforms normalised values in [0,1]
    into colours, with a normaliser that transforms the original values. The member `scalarMapper`
    """
    def __init__(self, norm_min, norm_max, cmap_points: List[Tuple[float, Any]], cmap_points_normalised=False):
        """
        :param norm_min: the value that shall be mapped to 0 in the normalised representation (any smaller values are also clipped to 0)
        :param norm_max: the value that shall be mapped to 1 in the normalised representation (any larger values are also clipped to 1)
        :param cmap_points: a list (of at least two) tuples (v, c) where v is the value and c is the colour associated with the value;
            any colour specification supported by matplotlib is admissible
        :param cmap_points_normalised: whether the values in `cmap_points` are already normalised
        """
        self.norm = matplotlib.colors.Normalize(vmin=norm_min, vmax=norm_max, clip=True)
        if not cmap_points_normalised:
            cmap_points = [(self.norm(v), c) for v, c in cmap_points]
        self.cmap = LinearSegmentedColormap.from_list(f"cmap{id(self)}", cmap_points)
        self.scalarMapper = matplotlib.cm.ScalarMappable(norm=self.norm, cmap=self.cmap)

    def get_color(self, value):
        rgba = self.scalarMapper.to_rgba(value)
        return '#%02x%02x%02x%02x' % tuple(int(v * 255) for v in rgba)


def plot_matrix(matrix: np.ndarray, title: str, xtick_labels: Sequence[str], ytick_labels: Sequence[str], xlabel: str,
        ylabel: str, normalize=True, figsize: Tuple[int, int] = (9, 9), title_add: str = None) -> matplotlib.figure.Figure:
    """
    :param matrix: matrix whose data to plot, where matrix[i, j] will be rendered at x=i, y=j
    :param title: the plot's title
    :param xtick_labels: the labels for the x-axis ticks
    :param ytick_labels: the labels for the y-axis ticks
    :param xlabel: the label for the x-axis
    :param ylabel: the label for the y-axis
    :param normalize: whether to normalise the matrix before plotting it (dividing each entry by the sum of all entries)
    :param figsize: an optional size of the figure to be created
    :param title_add: an optional second line to add to the title
    :return: the figure object
    """
    matrix = np.transpose(matrix)

    if title_add is not None:
        title += f"\n {title_add} "

    if normalize:
        matrix = matrix.astype('float') / matrix.sum()
    fig, ax = plt.subplots(figsize=figsize)
    fig.canvas.manager.set_window_title(title.replace("\n", " "))
    # We want to show all ticks...
    ax.set(xticks=np.arange(matrix.shape[1]),
        yticks=np.arange(matrix.shape[0]),
        # ... and label them with the respective list entries
        xticklabels=xtick_labels, yticklabels=ytick_labels,
        title=title,
        xlabel=xlabel,
        ylabel=ylabel)
    im = ax.imshow(matrix, interpolation='nearest', cmap=plt.cm.Blues)
    ax.figure.colorbar(im, ax=ax)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
        rotation_mode="anchor")

    # Loop over data dimensions and create text annotations.
    fmt = '.4f' if normalize else ('.2f' if matrix.dtype.kind == 'f' else 'd')
    thresh = matrix.max() / 2.
    for i in range(matrix.shape[0]):
        for j in range(matrix.shape[1]):
            ax.text(j, i, format(matrix[i, j], fmt),
                ha="center", va="center",
                color="white" if matrix[i, j] > thresh else "black")
    fig.tight_layout()
    return fig


TPlot = TypeVar("TPlot", bound="Plot")


class Plot:
    def __init__(self, draw: Callable[[plt.Axes], None] = None, name=None, ax: Optional[plt.Axes] = None):
        """
        :param draw: function which returns a matplotlib.Axes object to show
        :param name: name/number of the figure, which determines the window caption; it should be unique, as any plot
            with the same name will have its contents rendered in the same window. By default, figures are number
            sequentially.
        :param ax: the axes to draw to
        """
        if ax is not None:
            fig = None
        else:
            fig, ax = plt.subplots(num=name)
        self.fig: plt.Figure = fig
        self.ax: plt.Axes = ax
        if draw is not None:
            draw(ax)

    def xlabel(self: TPlot, label) -> TPlot:
        self.ax.set_xlabel(label)
        return self

    def ylabel(self: TPlot, label) -> TPlot:
        self.ax.set_ylabel(label)
        return self

    def title(self: TPlot, title: str) -> TPlot:
        self.ax.set_title(title)
        return self

    def xlim(self: TPlot, min_value, max_value) -> TPlot:
        self.ax.set_xlim(min_value, max_value)
        return self

    def ylim(self: TPlot, min_value, max_value) -> TPlot:
        self.ax.set_ylim(min_value, max_value)
        return self

    def save(self, path):
        log.info(f"Saving figure in {path}")
        self.fig.savefig(path)

    def xtick(self: TPlot, major=None, minor=None) -> TPlot:
        """
        Sets a tick on every integer multiple of the given base values.
        The major ticks are labelled, the minor ticks are not.

        :param major: the major tick base value
        :param minor: the minor tick base value
        :return: self
        """
        if major is not None:
            self.xtick_major(major)
        if minor is not None:
            self.xtick_minor(minor)
        return self

    def xtick_major(self: TPlot, base) -> TPlot:
        self.ax.xaxis.set_major_locator(plticker.MultipleLocator(base=base))
        return self

    def xtick_minor(self: TPlot, base) -> TPlot:
        self.ax.xaxis.set_minor_locator(plticker.MultipleLocator(base=base))
        return self

    def ytick_major(self: TPlot, base) -> TPlot:
        self.ax.yaxis.set_major_locator(plticker.MultipleLocator(base=base))
        return self


class ScatterPlot(Plot):
    N_MAX_TRANSPARENCY = 1000
    N_MIN_TRANSPARENCY = 100
    MAX_OPACITY = 0.5
    MIN_OPACITY = 0.05

    def __init__(self, x, y, c=None, c_base: Tuple[float, float, float] = (0, 0, 1), c_opacity=None, x_label=None,
            y_label=None, add_diagonal=False, **kwargs):
        """
        :param x: the x values; if has name (e.g. pd.Series), will be used as axis label
        :param y: the y values; if has name (e.g. pd.Series), will be used as axis label
        :param c: the colour specification; if None, compose from ``c_base`` and ``c_opacity``
        :param c_base: the base colour as (R, G, B) floats
        :param c_opacity: the opacity; if None, automatically determine from number of data points
        :param x_label:
        :param y_label:
        :param kwargs:
        """
        if c is None:
            if c_base is None:
                c_base = (0, 0, 1)
            if c_opacity is None:
                n = len(x)
                if n > self.N_MAX_TRANSPARENCY:
                    transparency = 1
                elif n < self.N_MIN_TRANSPARENCY:
                    transparency = 0
                else:
                    transparency = (n - self.N_MIN_TRANSPARENCY) / (self.N_MAX_TRANSPARENCY - self.N_MIN_TRANSPARENCY)
                c_opacity = self.MIN_OPACITY + (self.MAX_OPACITY - self.MIN_OPACITY) * (1-transparency)
            c = ((*c_base, c_opacity),)

        assert len(x) == len(y)
        if x_label is None and hasattr(x, "name"):
            x_label = x.name
        if y_label is None and hasattr(y, "name"):
            y_label = y.name

        def draw(ax):
            if x_label is not None:
                plt.xlabel(x_label)
            if x_label is not None:
                plt.ylabel(y_label)
            value_range = [min(min(x), min(y)), max(max(x), max(y))]
            plt.scatter(x, y, c=c, **kwargs)
            if add_diagonal:
                plt.plot(value_range, value_range, '-', lw=1, label="_not in legend", color="green", zorder=1)

        super().__init__(draw)


class HeatMapPlot(Plot):
    DEFAULT_CMAP_FACTORY = lambda num_points: LinearSegmentedColormap.from_list("whiteToRed",
        ((0, (1, 1, 1)), (1 / num_points, (1, 0.96, 0.96)), (1, (0.7, 0, 0))), num_points)

    def __init__(self, x, y, x_label=None, y_label=None, bins=60, cmap=None, common_range=True, diagonal=False,
            diagonal_color="green", **kwargs):
        """
        :param x: the x values
        :param y: the y values
        :param x_label: the x-axis label
        :param y_label: the y-axis label
        :param bins: the number of bins to use in each dimension
        :param cmap: the colour map to use for heat values (if None, use default)
        :param common_range: whether the heat map is to use a common rng for the x- and y-axes (set to False if x and y are completely
            different quantities; set to True use cases such as the evaluation of regression model quality)
        :param diagonal: whether to draw the diagonal line (useful for regression evaluation)
        :param diagonal_color: the colour to use for the diagonal line
        :param kwargs: parameters to pass on to plt.imshow
        """
        assert len(x) == len(y)
        if x_label is None and hasattr(x, "name"):
            x_label = x.name
        if y_label is None and hasattr(y, "name"):
            y_label = y.name

        def draw(ax):
            nonlocal cmap
            x_range = [min(x), max(x)]
            y_range = [min(y), max(y)]
            rng = [min(x_range[0], y_range[0]), max(x_range[1], y_range[1])]
            if common_range:
                x_range = y_range = rng
            if diagonal:
                plt.plot(rng, rng, '-', lw=0.75, label="_not in legend", color=diagonal_color, zorder=2)
            heatmap, _, _ = np.histogram2d(x, y, range=[x_range, y_range], bins=bins, density=False)
            extent = [x_range[0], x_range[1], y_range[0], y_range[1]]
            if cmap is None:
                cmap = HeatMapPlot.DEFAULT_CMAP_FACTORY(len(x))
            if x_label is not None:
                plt.xlabel(x_label)
            if y_label is not None:
                plt.ylabel(y_label)
            plt.imshow(heatmap.T, extent=extent, origin='lower', interpolation="none", cmap=cmap, zorder=1, aspect="auto", **kwargs)

        super().__init__(draw)


class HistogramPlot(Plot):
    def __init__(self, values, bins="auto", kde=False, cdf=False, cdf_complementary=False, cdf_secondary_axis=True,
            binwidth=None, stat="probability", xlabel=None,
            **kwargs):
        """
        :param values: the values to plot
        :param bins: a bin specification as understood by sns.histplot
        :param kde: whether to add a kernel density estimator
        :param cdf: whether to add a plot of the cumulative distribution function (cdf)
        :param cdf_complementary: whether to plot, if cdf is enabled, the complementary values
        :param cdf_secondary_axis: whether to use, if cdf is enabled, a secondary
        :param binwidth: the bin width; if None, inferred
        :param stat: the statistic to plot (as understood by sns.histplot)
        :param xlabel: the label for the x-axis
        :param kwargs: arguments to pass on to sns.histplot
        """

        def draw(ax):
            nonlocal cdf_secondary_axis
            sns.histplot(values, bins=bins, kde=kde, binwidth=binwidth, stat=stat, **kwargs)
            plt.ylabel(stat)
            if cdf:
                ecdf_stat = stat
                if ecdf_stat not in ("count", "proportion", "probability"):
                    ecdf_stat = "proportion"
                    cdf_secondary_axis = True
                cdf_ax: Optional[plt.Axes] = None
                cdf_ax_label = f"{ecdf_stat} (cdf)"
                if cdf_secondary_axis:
                    cdf_ax: plt.Axes = plt.twinx()
                    if stat in ("proportion", "probability"):
                        y_tick = 0.1
                    elif stat == "percent":
                        y_tick = 10
                    else:
                        y_tick = None
                    if y_tick is not None:
                        cdf_ax.yaxis.set_major_locator(plticker.MultipleLocator(base=y_tick))
                if cdf_complementary or ecdf_stat in ("count", "proportion", "probability"):
                    ecdf_stat = "proportion" if stat == "probability" else stat  # same semantics but "probability" not understood by ecdfplot
                    sns.ecdfplot(values, stat=ecdf_stat, complementary=cdf_complementary, color="orange", ax=cdf_ax)
                else:
                    sns.histplot(values, bins=100, stat=stat, element="poly", fill=False, cumulative=True, color="orange", ax=cdf_ax)
                if cdf_ax is not None:
                    cdf_ax.set_ylabel(cdf_ax_label)
            if xlabel is not None:
                self.xlabel(xlabel)

        super().__init__(draw)


class AverageSeriesLinePlot(Plot):
    """
    Plots the average of a collection of series or the averages of several collections of series,
    establishing a common index (the unification of all indices) for each collection via interpolation.
    The standard deviation is additionally shown as a shaded area around each line.
    """
    def __init__(self,
            series_collection: Union[List[pd.Series], Dict[str, List[pd.Series]]],
            interpolation: SeriesInterpolation,
            collection_name="collection",
            ax: Optional[plt.Axes] = None,
            hue_order=None, palette=None):
        """
        :param series_collection: either a list of series to average or a dictionary mapping the name of a collection
            to a list of series to average
        :param interpolation: the interpolation with which to obtain series values for the unified index of a collection of series
        :param collection_name: a name indicating what a key in `series_collection` refers to, which will appear in the legend
            for the case where more than one collection is passed
        :param ax: the axis to plot to; if None, create a new figure and axis
        :param hue_order: the hue order (for the case where there is more than one collection of series)
        :param palette: the colour palette to use
        """
        if isinstance(series_collection, dict):
            series_dict = series_collection
        else:
            series_dict = {"_": series_collection}

        series_list = next(iter(series_dict.values()))
        x_name = series_list[0].index.name or "x"
        y_name = series_list[0].name or "y"

        # build data frame with all series, interpolating each sub-collection
        dfs = []
        for name, series_list in series_dict.items():
            interpolated_series_list = interpolation.interpolate_all_with_combined_index(series_list)
            for series in interpolated_series_list:
                df = pd.DataFrame({y_name: series, x_name: series.index})
                df["series_id"] = id(series)
                df[collection_name] = name
                dfs.append(df)
        full_df = pd.concat(dfs, axis=0).reset_index(drop=True)

        def draw(ax):
            sns.lineplot(
                data=full_df,
                x=x_name,
                y=y_name,
                estimator="mean",
                hue=collection_name if len(series_dict) > 1 else None,
                hue_order=hue_order,
                palette=palette,
                ax=ax,
            )

        super().__init__(draw, ax=ax)



import logging
from typing import Optional

import pandas as pd
import xgboost

from . import InputOutputData
from .data import DataSplitter
from .sklearn.sklearn_base import AbstractSkLearnMultipleOneDimVectorRegressionModel, AbstractSkLearnVectorClassificationModel, \
    FeatureImportanceProviderSkLearnRegressionMultipleOneDim, FeatureImportanceProviderSkLearnClassification, ActualFitParams
from .util.pickle import setstate

log = logging.getLogger(__name__)


def is_xgboost_version_at_least(major: int, minor: Optional[int] = None, patch: Optional[int] = None):
    components = xgboost.__version__.split(".")
    for i, version in enumerate((major, minor, patch)):
        if version is not None:
            installed_version = int(components[i])
            if installed_version > version:
                return True
            if installed_version < version:
                return False
    return True


class XGBGradientBoostedVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel,
        FeatureImportanceProviderSkLearnRegressionMultipleOneDim):
    """
    XGBoost's regression model using gradient boosted trees
    """

    def __init__(self, random_state=42,
            early_stopping_rounds: Optional[int] = None,
            early_stopping_data_splitter: Optional[DataSplitter] = None,
            **model_args):
        """
        :param model_args: See https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBRegressor
        """
        super().__init__(xgboost.XGBRegressor, random_state=random_state, early_stopping_rounds=early_stopping_rounds,
            **model_args)
        self.is_early_stopping_enabled = early_stopping_rounds is not None
        self.early_stopping_data_splitter = early_stopping_data_splitter

    def __setstate__(self, state):
        setstate(XGBGradientBoostedVectorRegressionModel, self, state,
            new_default_properties=dict(
                is_early_stopping_enabled=False,
                early_stopping_data_splitter=None))

    def is_sample_weight_supported(self) -> bool:
        return True

    def _compute_actual_fit_params(self, inputs: pd.DataFrame, outputs: pd.DataFrame, weights: Optional[pd.Series] = None) -> ActualFitParams:
        kwargs = {}
        if self.is_early_stopping_enabled:
            data = InputOutputData(inputs, outputs, weights=weights)
            train_data, val_data = self.early_stopping_data_splitter.split(data)
            train_data: InputOutputData
            kwargs["eval_set"] = [(val_data.inputs, val_data.outputs)]
            inputs = train_data.inputs
            outputs = train_data.outputs
            weights = train_data.weights
            log.info(f"Early stopping enabled with validation set of size {len(val_data)}")
        params = super()._compute_actual_fit_params(inputs, outputs, weights=weights)
        params.kwargs.update(kwargs)
        return params


class XGBRandomForestVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel,
        FeatureImportanceProviderSkLearnRegressionMultipleOneDim):
    """
    XGBoost's random forest regression model
    """
    def __init__(self, random_state=42, **model_args):
        """
        :param model_args: See https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBRFRegressor
        """
        super().__init__(xgboost.XGBRFRegressor, random_state=random_state, **model_args)

    def is_sample_weight_supported(self) -> bool:
        return True


class XGBGradientBoostedVectorClassificationModel(AbstractSkLearnVectorClassificationModel, FeatureImportanceProviderSkLearnClassification):
    """
    XGBoost's classification model using gradient boosted trees
    """
    def __init__(self, random_state=42, use_balanced_class_weights=False, **model_args):
        """
        :param model_args: See https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBClassifier
        """
        use_label_encoding = is_xgboost_version_at_least(1, 6)
        super().__init__(xgboost.XGBClassifier, random_state=random_state, use_balanced_class_weights=use_balanced_class_weights,
            use_label_encoding=use_label_encoding, **model_args)

    def is_sample_weight_supported(self) -> bool:
        return True


class XGBRandomForestVectorClassificationModel(AbstractSkLearnVectorClassificationModel, FeatureImportanceProviderSkLearnClassification):
    """
    XGBoost's random forest classification model
    """
    def __init__(self, random_state=42, use_balanced_class_weights=False, **model_args):
        """
        :param model_args: See https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBRFClassifier
        """
        use_label_encoding = is_xgboost_version_at_least(1, 6)
        super().__init__(xgboost.XGBRFClassifier, random_state=random_state, use_balanced_class_weights=use_balanced_class_weights,
            use_label_encoding=use_label_encoding, **model_args)

    def is_sample_weight_supported(self) -> bool:
        return True



"""
Utility functions and classes for geographic coordinates
"""

import math
from typing import Tuple, Iterable

import numpy as np
import pandas as pd

from ..util.string import ToStringMixin

EARTH_RADIUS = 6371000
EARTH_CIRCUMFERENCE = 2 * math.pi * EARTH_RADIUS
LATITUDE_PER_METRE = 360.0 / EARTH_CIRCUMFERENCE


def longitude_per_m(latitude):
    return LATITUDE_PER_METRE / math.cos(math.radians(latitude))


def latitude_per_m():
    return LATITUDE_PER_METRE


def approximate_squared_distance(p1: Tuple[float, float], p2: Tuple[float, float]):
    """
    :param p1: a tuple (latitude, longitude)
    :param p2: a tuple (latitude, longitude)
    :return: the approximate squared distance (in m²) between p1 and p2
    """
    lat_per_m = latitude_per_m()
    p1lat, p1lon = p1
    p2lat, p2lon = p2
    lon_per_m = longitude_per_m((p1lat + p2lat) / 2)
    dx = (p2lon - p1lon) / lon_per_m
    dy = (p2lat - p1lat) / lat_per_m
    return dx * dx + dy * dy


def closest_point_on_segment(search_pos: Tuple[float, float], segPoint1: Tuple[float, float], segPoint2: Tuple[float, float]):
    """
    Gets the point on the line segment connecting segPoint1 and segPoint2 that is closest to searchPos

    :param search_pos: the position for which to search for the closest point on the line segment
    :param segPoint1: the first point defining the line segment on which to search
    :param segPoint2: the second point defining the line segment on which to search
    :return: the closest point, which is on the line connecting segPoint1 and segPoint2 (and may be one of the two points)
    """
    seg1lat, seg1lon = segPoint1
    seg2lat, seg2lon = segPoint2
    srchlat, srchlon = search_pos
    lat_per_m = latitude_per_m()
    lon_per_m = longitude_per_m(srchlat)
    sp1x = (seg1lon - srchlon) / lon_per_m
    sp1y = (seg1lat - srchlat) / lat_per_m
    sp2x = (seg2lon - srchlon) / lon_per_m
    sp2y = (seg2lat - srchlat) / lat_per_m
    vx = sp2x - sp1x
    vy = sp2y - sp1y
    c1 = -vx * sp1x - vy * sp1y
    if c1 <= 0:
        return segPoint1
    c2 = vx * vx + vy * vy
    if c2 <= c1:
        return segPoint2
    b = 0 if c2 == 0 else c1 / c2
    lon = seg1lon + b * vx * lon_per_m
    lat = seg1lat + b * vy * lat_per_m
    return [lat, lon]


def orientation(p1: Tuple[float, float], p2: Tuple[float, float]) -> float:
    """
    Gets the orientation angle for the vector from p1 to p2

    :param p1: a (lat, lon) pair
    :param p2: a (lat, lon) pair
    :return: the orientation angle in rad
    """
    p1_lat, p1_lon = p1
    p2_lat, p2_lon = p2
    center_lat = (p1_lat + p2_lat) / 2
    dx = (p2_lon - p1_lon) / longitude_per_m(center_lat)
    dy = (p2_lat - p1_lat) / latitude_per_m()
    return math.atan2(dy, dx)


def abs_angle_difference(a1: float, a2: float) -> float:
    """
    Computes the absolute angle difference in ]-pi, pi] between two angles

    :param a1: an angle in rad
    :param a2: an angle in rad
    :return: the difference in rad
    """
    d = a1 - a2
    while d > math.pi:
        d -= 2*math.pi
    while d <= -math.pi:
        d += 2*math.pi
    return abs(d)


def closest_point_on_polyline(search_pos, polyline, search_orientation_angle=None, max_angle_difference=0) \
        -> Tuple[Tuple[float, float], float, int]:
    """
    Gets the point on the given polyline that is closest to the given search position along with the
    distance (in metres) to the polyline

    :param search_pos: a (lat, lon) pair indicating the position for which to find the closest math on the polyline
    :param polyline: list of (lat, lon) pairs that make up the polyline on which to search
    :param search_orientation_angle: if not None, defines the orientation with which to compute angle differences
        (if maxAngleDifference > 0)
    :param max_angle_difference: the maximum absolute angle difference (in rad) that is admissible (between the orientation of the
        respective line segment and the orientation given in searchOrientationAngle)
    :return: a tuple (opt_point, opt_dist, opt_segment_start_idx) where
        opt_point is the closest point (with admissible orientation - or None if there is none),
        opt_dist is the distance from the polyline to the closest point,
        opt_segment_start_idx is the index of the first point of the segment on the polyline for which the closest point was found
    """
    if len(polyline) < 2:
        raise Exception("Polyline must consist of at least two points")
    opt_segment_start_idx = None
    opt_point = None
    opt_sq_dist = None
    for i in range(len(polyline)-1):
        if max_angle_difference > 0:
            orientation_angle = orientation(polyline[i], polyline[i+1])
            ang_diff = abs_angle_difference(orientation_angle, search_orientation_angle)
            if ang_diff > max_angle_difference:
                continue
        opt_seg_point = closest_point_on_segment(search_pos, polyline[i], polyline[i + 1])
        sq_dist = approximate_squared_distance(search_pos, opt_seg_point)
        if opt_sq_dist is None or sq_dist < opt_sq_dist:
            opt_point = opt_seg_point
            opt_sq_dist = sq_dist
            opt_segment_start_idx = i
    return opt_point, math.sqrt(opt_sq_dist), opt_segment_start_idx


class GeoCoord(ToStringMixin):
    """
    Represents geographic coordinates (WGS84)
    """
    def __init__(self, lat: float, lon: float):
        self.lat = lat
        self.lon = lon

    def latlon(self):
        return self.lat, self.lon

    def distance_to(self, gps_position: 'GeoCoord'):
        return math.sqrt(self.squared_distance_to(gps_position))

    def squared_distance_to(self, gps_position: 'GeoCoord'):
        return approximate_squared_distance(self.latlon(), gps_position.latlon())

    def local_coords(self, lcs):
        return lcs.get_local_coords(self.lat, self.lon)

    @classmethod
    def mean_coord(cls, geo_coords: Iterable["GeoCoord"]):
        mean_lat = np.mean([c.lat for c in geo_coords])
        mean_lon = np.mean([c.lon for c in geo_coords])
        # noinspection PyTypeChecker
        return GeoCoord(mean_lat, mean_lon)


class GpsTracePoint(GeoCoord):
    def __init__(self, lat, lon, time: pd.Timestamp):
        super().__init__(lat, lon)
        self.time = time


class GeoRect:
    def __init__(self, min_lat: float, min_lon: float, max_lat: float, max_lon: float):
        if max_lat < min_lat or max_lon < min_lon:
            raise ValueError()
        self.minLat = min_lat
        self.minLon = min_lon
        self.maxLat = max_lat
        self.maxLon = max_lon

    @staticmethod
    def from_circle(centre_lat, centre_lon, radius_m):
        """Creates the bounding rectangle for the given circular area"""
        from .local_coords import LocalCoordinateSystem
        lcs = LocalCoordinateSystem(centre_lat, centre_lon)
        min_lat, min_lon = lcs.get_lat_lon(-radius_m, -radius_m)
        max_lat, max_lon = lcs.get_lat_lon(radius_m, radius_m)
        return GeoRect(min_lat, min_lon, max_lat, max_lon)



from abc import ABC, abstractmethod
from typing import Generic, TypeVar, List, Union, Dict, Sequence, Optional, Tuple, Callable

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from ...util.pickle import setstate
from ...util.plot import ScatterPlot, HistogramPlot, Plot, HeatMapPlot
from ...util.string import ToStringMixin, dict_string
from ...vector_model import VectorModel

# Note: in the 2020.2 version of PyCharm passing strings to bound is highlighted as error
# It does not cause runtime errors and the static type checker ignores the bound anyway, so it does not matter for now.
# However, this might cause problems with type checking in the future. Therefore, I moved the definition of TEvalStats
# below the definition of EvalStats. Unfortunately, the dependency in generics between EvalStats and Metric
# does not allow to define both, TMetric and TEvalStats, properly. For now we have to leave it with the bound as string
# and hope for the best in the future
TMetric = TypeVar("TMetric", bound="Metric")
TVectorModel = TypeVar("TVectorModel", bound=VectorModel)

Array = Union[np.ndarray, pd.Series, list]
PredictionArray = Union[np.ndarray, pd.Series, pd.DataFrame, list]


class EvalStats(Generic[TMetric], ToStringMixin):
    def __init__(self, metrics: List[TMetric], additional_metrics: List[TMetric] = None):
        if len(metrics) == 0:
            raise ValueError("No metrics provided")
        self.metrics = metrics
        # Implementations of EvalStats will typically provide default metrics, therefore we include
        # the possibility for passing additional metrics here
        if additional_metrics is not None:
            self.metrics = self.metrics + additional_metrics
        self.name = None

    def set_name(self, name: str):
        self.name = name

    def add_metric(self, metric: TMetric):
        self.metrics.append(metric)

    def compute_metric_value(self, metric: TMetric) -> float:
        return metric.compute_value_for_eval_stats(self)

    def metrics_dict(self) -> Dict[str, float]:
        """
        Computes all metrics

        :return: a dictionary mapping metric names to values
        """
        d = {}
        for metric in self.metrics:
            d[metric.name] = self.compute_metric_value(metric)
        return d

    def get_all(self) -> Dict[str, float]:
        """Alias for metricsDict; may be deprecated in the future"""
        return self.metrics_dict()

    def _tostring_object_info(self) -> str:
        return dict_string(self.metrics_dict())


TEvalStats = TypeVar("TEvalStats", bound=EvalStats)


class Metric(Generic[TEvalStats], ABC):
    name: str

    def __init__(self, name: str = None, bounds: Optional[Tuple[float, float]] = None):
        """
        :param name: the name of the metric; if None use the class' name attribute
        :param bounds: the minimum and maximum values the metric can take on (or None if the bounds are not specified)
        """
        # this raises an attribute error if a subclass does not specify a name as a static attribute nor as parameter
        self.name = name if name is not None else self.__class__.name
        self.bounds = bounds

    @abstractmethod
    def compute_value_for_eval_stats(self, eval_stats: TEvalStats) -> float:
        pass

    def get_paired_metrics(self) -> List[TMetric]:
        """
        Gets a list of metrics that should be considered together with this metric (e.g. for paired visualisations/plots).
        The direction of the pairing should be such that if this metric is "x", the other is "y" for x-y type visualisations.

        :return: a list of metrics
        """
        return []

    def has_finite_bounds(self) -> bool:
        return self.bounds is not None and not any((np.isinf(x) for x in self.bounds))


class EvalStatsCollection(Generic[TEvalStats, TMetric], ABC):
    def __init__(self, eval_stats_list: List[TEvalStats]):
        self.statsList = eval_stats_list
        metric_names_set = None
        metrics_list = []
        for es in eval_stats_list:
            metrics = es.metrics_dict()
            current_metric_names_set = set(metrics.keys())
            if metric_names_set is None:
                metric_names_set = current_metric_names_set
            else:
                if metric_names_set != current_metric_names_set:
                    raise Exception(f"Inconsistent set of metrics in evaluation stats collection: "
                                    f"Got {metric_names_set} for one instance, {current_metric_names_set} for another")
            metrics_list.append(metrics)
        metric_names = sorted(metrics_list[0].keys())
        self._valuesByMetricName = {metric: [d[metric] for d in metrics_list] for metric in metric_names}
        self._metrics: List[TMetric] = eval_stats_list[0].metrics

    def get_values(self, metric_name: str):
        return self._valuesByMetricName[metric_name]

    def get_metric_names(self) -> List[str]:
        return list(self._valuesByMetricName.keys())

    def get_metrics(self) -> List[TMetric]:
        return self._metrics

    def get_metric_by_name(self, name: str) -> Optional[TMetric]:
        for m in self._metrics:
            if m.name == name:
                return m
        return None

    def has_metric(self, metric: Union[Metric, str]) -> bool:
        if type(metric) != str:
            metric = metric.name
        return metric in self._valuesByMetricName

    def agg_metrics_dict(self, agg_fns=(np.mean, np.std)) -> Dict[str, float]:
        agg = {}
        for metric, values in self._valuesByMetricName.items():
            for agg_fn in agg_fns:
                agg[f"{agg_fn.__name__}[{metric}]"] = float(agg_fn(values))
        return agg

    def mean_metrics_dict(self) -> Dict[str, float]:
        metrics = {metric: np.mean(values) for (metric, values) in self._valuesByMetricName.items()}
        return metrics

    def plot_distribution(self, metric_name: str, subtitle: Optional[str] = None, bins=None, kde=False, cdf=False,
            cdf_complementary=False, stat="proportion", **kwargs) -> plt.Figure:
        """
        Plots the distribution of a metric as a histogram

        :param metric_name: name of the metric for which to plot the distribution (histogram) across evaluations
        :param subtitle: the subtitle to add, if any
        :param bins: the histogram bins (number of bins or boundaries); metrics bounds will be used to define the x limits.
            If None, use 'auto' bins
        :param kde: whether to add a kernel density estimator plot
        :param cdf: whether to add the cumulative distribution function (cdf)
        :param cdf_complementary: whether to plot, if ``cdf`` is True, the complementary cdf instead of the regular cdf
        :param stat: the statistic to compute for each bin ('percent', 'probability'='proportion', 'count', 'frequency' or 'density'),
            y-axis value
        :param kwargs: additional parameters to pass to seaborn.histplot (see https://seaborn.pydata.org/generated/seaborn.histplot.html)
        :return: the plot
        """
        # define bins based on metric bounds where available
        x_tick = None
        if bins is None or type(bins) == int:
            metric = self.get_metric_by_name(metric_name)
            if metric.bounds == (0, 1):
                x_tick = 0.1
                if bins is None:
                    num_bins = 10 if cdf else 20
                else:
                    num_bins = bins
                bins = np.linspace(0, 1, num_bins+1)
            else:
                bins = "auto"

        values = self._valuesByMetricName[metric_name]
        title = metric_name
        if subtitle is not None:
            title += "\n" + subtitle
        plot = HistogramPlot(values, bins=bins, stat=stat, kde=kde, cdf=cdf, cdf_complementary=cdf_complementary, **kwargs).title(title)
        if x_tick is not None:
            plot.xtick_major(x_tick)
        return plot.fig

    def _plot_xy(self, metric_name_x, metric_name_y, plot_factory: Callable[[Sequence, Sequence], Plot], adjust_bounds: bool) -> plt.Figure:
        def axlim(bounds):
            min_value, max_value = bounds
            diff = max_value - min_value
            return (min_value - 0.05 * diff, max_value + 0.05 * diff)

        x = self._valuesByMetricName[metric_name_x]
        y = self._valuesByMetricName[metric_name_y]
        plot = plot_factory(x, y)
        plot.xlabel(metric_name_x)
        plot.ylabel(metric_name_y)
        mx = self.get_metric_by_name(metric_name_x)
        if adjust_bounds and mx.has_finite_bounds():
            plot.xlim(*axlim(mx.bounds))
        my = self.get_metric_by_name(metric_name_y)
        if adjust_bounds and my.has_finite_bounds():
            plot.ylim(*axlim(my.bounds))
        return plot.fig

    def plot_scatter(self, metric_name_x: str, metric_name_y: str) -> plt.Figure:
        return self._plot_xy(metric_name_x, metric_name_y, ScatterPlot, adjust_bounds=True)

    def plot_heat_map(self, metric_name_x: str, metric_name_y: str) -> plt.Figure:
        return self._plot_xy(metric_name_x, metric_name_y, HeatMapPlot, adjust_bounds=False)

    def to_data_frame(self) -> pd.DataFrame:
        """
        :return: a DataFrame with the evaluation metrics from all contained EvalStats objects;
            the EvalStats' name field being used as the index if it is set
        """
        data = dict(self._valuesByMetricName)
        index = [stats.name for stats in self.statsList]
        if len([n for n in index if n is not None]) == 0:
            index = None
        return pd.DataFrame(data, index=index)

    def get_global_stats(self) -> TEvalStats:
        """
        Alias for `getCombinedEvalStats`
        """
        return self.get_combined_eval_stats()

    @abstractmethod
    def get_combined_eval_stats(self) -> TEvalStats:
        """
        :return: an EvalStats object that combines the data from all contained EvalStats objects
        """
        pass

    def __str__(self):
        return f"{self.__class__.__name__}[" + \
               ", ".join([f"{key}={self.agg_metrics_dict()[key]:.4f}" for key in self._valuesByMetricName]) + "]"


class PredictionEvalStats(EvalStats[TMetric], ABC):
    """
    Collects data for the evaluation of predicted values (including multi-dimensional predictions)
    and computes corresponding metrics
    """
    def __init__(self, y_predicted: Optional[PredictionArray], y_true: Optional[PredictionArray],
            metrics: List[TMetric], additional_metrics: List[TMetric] = None,
            weights: Optional[Array] = None):
        """
        :param y_predicted: sequence of predicted values, or, in case of multi-dimensional predictions, either a data frame with
            one column per dimension or a nested sequence of values
        :param y_true: sequence of ground truth labels of same shape as y_predicted
        :param metrics: list of metrics to be computed on the provided data
        :param additional_metrics: the metrics to additionally compute. This should only be provided if metrics is None
        :param weights: weights for each data point contained in `y_predicted` and `y_true`
        """
        self.y_true = []
        self.y_predicted = []
        self.weights: Optional[List[float]] = None
        self.y_true_multidim = None
        self.y_predicted_multidim = None
        if y_predicted is not None:
            self.add_all(y_predicted=y_predicted, y_true=y_true, weights=weights)
        super().__init__(metrics, additional_metrics=additional_metrics)

    def __setstate__(self, state):
        return setstate(PredictionEvalStats, self, state, new_optional_properties=["weights"])

    def add(self, y_predicted, y_true, weight: Optional[float] = None) -> None:
        """
        Adds a single pair of values to the evaluation

        :param y_predicted: the value predicted by the model
        :param y_true: the true value
        """
        self.y_true.append(y_true)
        self.y_predicted.append(y_predicted)
        if weight is not None:
            if self.weights is None:
                self.weights = []
            self.weights.append(weight)

    def add_all(self, y_predicted: PredictionArray, y_true: PredictionArray, weights: Optional[Array] = None) -> None:
        """
        :param y_predicted: sequence of predicted values, or, in case of multi-dimensional predictions, either a data frame with
            one column per dimension or a nested sequence of values
        :param y_true: sequence of ground truth labels of same shape as y_predicted
        :param weights: optional weights of data points
        """
        def is_sequence(x):
            return isinstance(x, pd.Series) or isinstance(x, list) or isinstance(x, np.ndarray)

        if is_sequence(y_predicted) and is_sequence(y_true):
            a, b = len(y_predicted), len(y_true)
            if a != b:
                raise Exception(f"Lengths differ (predicted {a}, truth {b})")
            if a > 0:
                first_item = y_predicted.iloc[0] if isinstance(y_predicted, pd.Series) else y_predicted[0]
                is_nested_sequence = is_sequence(first_item)
                if is_nested_sequence:
                    for y_true_i, y_predicted_i in zip(y_true, y_predicted):
                        self.add_all(y_predicted=y_predicted_i, y_true=y_true_i)
                else:
                    self.y_true.extend(y_true)
                    self.y_predicted.extend(y_predicted)
        elif isinstance(y_predicted, pd.DataFrame) and isinstance(y_true, pd.DataFrame):
            # keep track of multidimensional data (to be used later in getEvalStatsCollection)
            y_predicted_multidim = y_predicted.values
            y_true_multidim = y_true.values
            dim = y_predicted_multidim.shape[1]
            if dim != y_true_multidim.shape[1]:
                raise Exception("Dimension mismatch")
            if self.y_true_multidim is None:
                self.y_predicted_multidim = [[] for _ in range(dim)]
                self.y_true_multidim = [[] for _ in range(dim)]
            if len(self.y_predicted_multidim) != dim:
                raise Exception("Dimension mismatch")
            for i in range(dim):
                self.y_predicted_multidim[i].extend(y_predicted_multidim[:, i])
                self.y_true_multidim[i].extend(y_true_multidim[:, i])
            # convert to flat data for this stats object
            y_predicted = y_predicted_multidim.reshape(-1)
            y_true = y_true_multidim.reshape(-1)
            self.y_true.extend(y_true)
            self.y_predicted.extend(y_predicted)
        else:
            raise Exception(f"Unhandled data types: {type(y_predicted)}, {type(y_true)}")

        if weights is not None:
            if self.weights is None:
                self.weights = []
            assert len(weights) == len(self.y_predicted) - len(self.weights), "Length of weights does not match"
            self.weights.extend(weights)

    def _tostring_object_info(self) -> str:
        return f"{super()._tostring_object_info()}, N={len(self.y_predicted)}"


def mean_stats(eval_stats_list: Sequence[EvalStats]) -> Dict[str, float]:
    """
    For a list of EvalStats objects compute the mean values of all metrics in a dictionary.
    Assumes that all provided EvalStats have the same metrics
    """
    dicts = [s.metrics_dict() for s in eval_stats_list]
    metrics = dicts[0].keys()
    return {m: np.mean([d[m] for d in dicts]) for m in metrics}


class EvalStatsPlot(Generic[TEvalStats], ABC):
    @abstractmethod
    def create_figure(self, eval_stats: TEvalStats, subtitle: str) -> Optional[plt.Figure]:
        """
        :param eval_stats: the evaluation stats from which to generate the plot
        :param subtitle: the plot's subtitle
        :return: the figure or None if this plot is not applicable/cannot be created
        """
        pass

    def is_applicable(self, eval_stats: TEvalStats) -> bool:
        return True



"""
This module defines base classes for models that use pandas.DataFrames for inputs and outputs, where each data frame row represents
a single model input or output. Since every row contains a vector of data (one-dimensional array), we refer to them as vector-based
models. Hence the name of the module and of the central base class :class:`VectorModel`.
"""

import logging
import typing
from abc import ABC, abstractmethod
from typing import List, Any, Optional, Union, Type, Dict

import numpy as np
import pandas as pd

from .util.deprecation import deprecated
from .data import InputOutputData
from .data_transformation import DataFrameTransformer, DataFrameTransformerChain, InvertibleDataFrameTransformer
from .featuregen import FeatureGenerator, FeatureCollector
from .util import mark_used, kwarg_if_not_none
from .util.cache import PickleLoadSaveMixin
from .util.logging import StopWatch
from .util.pickle import setstate, getstate
from .util.sequences import get_first_duplicate
from .util.string import ToStringMixin

mark_used(InputOutputData)  # for backward compatibility

log = logging.getLogger(__name__)
TVectorModelBase = typing.TypeVar("TVectorModelBase", bound="VectorModelBase")
TVectorModel = typing.TypeVar("TVectorModel", bound="VectorModel")
TVectorRegressionModel = typing.TypeVar("TVectorRegressionModel", bound="VectorRegressionModel")


class VectorModelBase(ABC, ToStringMixin):
    """
    Base class for vector models, which defines the fundamental prediction interface.
    A vector model takes data frames as input, where each row represents a vector of information.
    """
    def __init__(self):
        self._name = None

    @abstractmethod
    def predict(self, x: pd.DataFrame) -> pd.DataFrame:
        pass

    @abstractmethod
    def is_regression_model(self) -> bool:
        pass

    @abstractmethod
    def get_predicted_variable_names(self) -> list:
        pass

    def with_name(self: TVectorModelBase, name: str) -> TVectorModelBase:
        """
        Sets the model's name.

        :param name: the name
        :return: self
        """
        self.set_name(name)
        return self

    def set_name(self, name):
        self._name = name

    def get_name(self):
        if self._name is None:
            return "unnamed-%s-%x" % (self.__class__.__name__, id(self))
        return self._name


class TrainingContext:
    """
    Contains context information for an ongoing training process
    """
    def __init__(self, original_input: pd.DataFrame, original_output: pd.DataFrame):
        self.original_input = original_input
        self.original_output = original_output


class VectorModel(VectorModelBase, PickleLoadSaveMixin, ABC):
    """
    Represents a model which uses data frames as inputs and outputs whose rows define individual data points.
    Every data frame row represents a vector of information (one-dimensional array), hence the name of the model.
    Note that the vectors in question are not necessarily vectors in the mathematical sense, as the information in each cell is not
    required to be numeric or uniform but can be arbitrarily complex.
    """
    TOSTRING_INCLUDE_PREPROCESSORS = True
    _TRANSIENT_MEMBERS = ["_trainingContext"]
    _RENAMED_MEMBERS = {
        "checkInputColumns": "_checkInputColumns",
        "_inputTransformerChain": "_featureTransformerChain"
    }

    def __init__(self, check_input_columns=True):
        """
        :param check_input_columns: whether to check if the input column list (that is fed to the underlying model, i.e. after feature
            generation) during inference coincides with the input column list that was observed during training.
            This should be disabled if feature generation is not performed by the model itself, e.g. in meta-models
            such as ensemble models.
        """
        super().__init__()
        self._featureGenerator: Optional[FeatureGenerator] = None
        self._rawInputTransformerChain = DataFrameTransformerChain()
        self._featureTransformerChain = DataFrameTransformerChain()
        self._isFitted = False  # Note: this keeps track only of the actual model being fitted, not the pre/postprocessors
        self._predictedVariableNames: Optional[list] = None
        self._modelInputVariableNames: Optional[list] = None
        self._checkInputColumns = check_input_columns

        # transient members
        self._trainingContext: Optional[TrainingContext] = None

    def __getstate__(self):
        return getstate(VectorModel, self, transient_properties=self._TRANSIENT_MEMBERS)

    def __setstate__(self, state):
        for m in VectorModel._TRANSIENT_MEMBERS:
            state[m] = None
        setstate(VectorModel, self, state, renamed_properties=self._RENAMED_MEMBERS,
            new_default_properties={"_rawInputTransformerChain": DataFrameTransformerChain()})

    def _tostring_exclude_private(self) -> bool:
        return True

    def _tostring_exclude_exceptions(self) -> List[str]:
        e = super()._tostring_exclude_exceptions()
        if self.TOSTRING_INCLUDE_PREPROCESSORS:
            e += ["_rawInputTransformerChain", "_featureGenerator", "_featureTransformerChain"]
        return e

    def _tostring_additional_entries(self) -> Dict[str, Any]:
        d = super()._tostring_additional_entries()
        if self._featureGenerator is not None:
            d["featureGeneratorNames"] = self._featureGenerator.get_names()
        if self._name is not None:
            d["name"] = self._name
        return d

    def with_raw_input_transformers(self: TVectorModel,
            *transformers: Union[DataFrameTransformer, List[DataFrameTransformer]]) -> TVectorModel:
        """
        Makes the model use the given transformers (removing previously set raw input transformers, if any), which
        are to be applied to the raw input data frame (prior to feature generation).

        :param transformers: :class:`DataFrameTransformer` instances to use (in sequence) for the transformation of inputs
        :return: self
        """
        self._rawInputTransformerChain = DataFrameTransformerChain(*transformers)
        return self

    def with_feature_transformers(self: TVectorModel, *transformers: Union[DataFrameTransformer, List[DataFrameTransformer]],
            add=False) -> TVectorModel:
        """
        Makes the model use the given transformers
        which are to be applied to the data frames generated by feature generators.
        (If the model does not use feature generators, the transformers will be applied to
        whatever is produced by the raw input transformers or, if there are none, the original raw
        input data frame).

        :param transformers: :class:`DataFrameTransformer` instances to use (in sequence) for the transformation of features
        :param add: whether to add the transformers to the existing transformers rather than replacing them
        :return: self
        """
        if not add:
            self._featureTransformerChain = DataFrameTransformerChain(*transformers)
        else:
            for t in transformers:
                self._featureTransformerChain.append(t)
        return self

    @deprecated("Use with_feature_transformers instead; this method will be removed in a future sensAI release.")
    def with_input_transformers(self: TVectorModel,
            *input_transformers: Union[DataFrameTransformer, List[DataFrameTransformer]]) -> TVectorModel:
        """
        Makes the model use the given feature transformers (removing previously set transformers, if any),
        i.e. it transforms the data frame that is generated by the feature generators (if any).

        :param input_transformers: :class:`DataFrameTransformer` instances to use (in sequence) for the transformation of inputs
        :return: self
        """
        return self.with_feature_transformers(*input_transformers)

    def with_feature_generator(self: TVectorModel, feature_generator: Optional[FeatureGenerator]) -> TVectorModel:
        """
        Makes the model use the given feature generator in order to obtain the model inputs.
        If the model shall use more than one feature generator, pass a :class:`MultiFeatureGenerator` which combines them or
        use the perhaps more convenient :class:`FeatureCollector` in conjunction with :meth:`withFeatureCollector`.

        Note: Feature computation takes place before input transformation.

        :param feature_generator: the feature generator to use for input computation
        :return: self
        """
        self._featureGenerator = feature_generator
        return self

    def with_feature_collector(self: TVectorModel, feature_collector: FeatureCollector,
            shared: bool = False) -> TVectorModel:
        """
        Makes the model use a multi-feature generator obtained from the given collector
        in order compute the underlying model's input from the data frame that is given.
        Overrides any feature generator previously passed to :meth:`withFeatureGenerator` (if any).

        Note: Feature generation takes place before feature transformation.

        :param feature_collector: the feature collector from which to obtain the multi-feature generator
        :param shared: whether the given feature collector is shared between models (i.e. whether
            the same instance is passed to multiple models).
            Passing `shared=True` ensures that models using the same collector do not end up
            using the same multi-feature collector instance and instead receive an independent instance.
        :return: self
        """
        if shared:
            self._featureGenerator = feature_collector.create_multi_feature_generator()
        else:
            self._featureGenerator = feature_collector.get_multi_feature_generator()
        return self

    def _pre_processors_are_fitted(self):
        result = self._rawInputTransformerChain.is_fitted() and self._featureTransformerChain.is_fitted()
        if self.get_feature_generator() is not None:
            result = result and self.get_feature_generator().is_fitted()
        return result

    def is_fitted(self):
        """
        :return: True if the model has been fitted, False otherwise
        """
        if not self._is_underlying_model_fitted():
            return False
        if not self._pre_processors_are_fitted():
            return False
        return True

    def _is_underlying_model_fitted(self):
        underlying_model_is_fitted = not self._underlying_model_requires_fitting() or self._isFitted
        return underlying_model_is_fitted

    def _check_model_input_columns(self, model_input: pd.DataFrame):
        if self._checkInputColumns and list(model_input.columns) != self._modelInputVariableNames:
            raise Exception(f"Inadmissible input data frame: "
                            f"expected columns {self._modelInputVariableNames}, got {list(model_input.columns)}")

    def compute_model_inputs(self, x: pd.DataFrame):
        """
        Applies feature generators and input transformers (if any) to generate from an input data frame the input for the
        underlying model

        :param x: the input data frame, to which input preprocessing is to be applied
        :return: the input data frame that serves as input for the underlying model
        """
        return self._compute_model_inputs(x)

    def _compute_model_inputs(self, x: pd.DataFrame, y: pd.DataFrame = None, fit=False) -> pd.DataFrame:
        """
        :param x: the input data frame
        :param y: the output data frame (when training); only has to be provided if ``fit=True`` and preprocessors require outputs
            for fitting
        :param fit: if True, preprocessors will be fitted before being applied to ``X``
        :return:
        """
        if fit:
            x = self._rawInputTransformerChain.fit_apply(x)
            if self._featureGenerator is not None:
                x = self._featureGenerator.fit_generate(x, y, self)
            x = self._featureTransformerChain.fit_apply(x)
        else:
            x = self._rawInputTransformerChain.apply(x)
            if self._featureGenerator is not None:
                x = self._featureGenerator.generate(x, self)
            x = self._featureTransformerChain.apply(x)
        return x

    def _compute_model_outputs(self, y: pd.DataFrame) -> pd.DataFrame:
        return y

    def compute_model_outputs(self, y: pd.DataFrame) -> pd.DataFrame:
        return self._compute_model_outputs(y)

    def predict(self, x: pd.DataFrame) -> pd.DataFrame:
        """
        Applies the model to the given input data frame

        :param x: the input data frame
        :return: the model outputs in the form of a data frame whose index corresponds to the index of ``x``
        """
        if not self.is_fitted():
            raise Exception(f"Calling predict with unfitted model {self} "
                            f"(isUnderlyingModelFitted={self._is_underlying_model_fitted()}, "
                            f"preProcessorsAreFitted={self._pre_processors_are_fitted()})")
        x = self._compute_model_inputs(x)
        self._check_model_input_columns(x)
        y = self._predict(x)
        return self._create_output_data_frame(y, x.index)

    def _create_output_data_frame(self, y: Union[pd.DataFrame, list], index):
        if isinstance(y, pd.DataFrame):
            # make sure the data frame has the right index
            y.index = index
            return y
        else:
            predicted_columns = self.get_predicted_variable_names()
            if len(predicted_columns) != 1:
                raise ValueError(f"_predict must return a DataFrame as there are multiple predicted columns; got {type(y)}")
            return pd.DataFrame(pd.Series(y, name=predicted_columns[0], index=index))

    @abstractmethod
    def _predict(self, x: pd.DataFrame) -> Union[pd.DataFrame, list]:
        """
        :param x: the input data frame
        :return: the output data frame, or, for the case where a single column is to be predicted, the list of values for that column
        """
        pass

    def _underlying_model_requires_fitting(self) -> bool:
        """
        Designed to be overridden for rule-based models.

        :return: True iff the underlying model requires fitting
        """
        return True

    def _fit_preprocessors(self, x: pd.DataFrame, y: pd.DataFrame = None):
        self._rawInputTransformerChain.fit(x)
        # no need for fitGenerate if chain is empty
        if self._featureGenerator is not None:
            if len(self._featureTransformerChain) == 0:
                self._featureGenerator.fit(x, y)
            else:
                x = self._featureGenerator.fit_generate(x, y, self)
        self._featureTransformerChain.fit(x)

    def fit_input_output_data(self, io_data: InputOutputData, fit_preprocessors=True, fit_model=True):
        """
        Fits the model using the given data

        :param io_data: the input/output data
        :param fit_preprocessors: whether the model's preprocessors (feature generators and data frame transformers) shall be fitted
        :param fit_model: whether the model itself shall be fitted
        """
        self.fit(io_data.inputs, io_data.outputs, weights=io_data.weights, fit_preprocessors=fit_preprocessors, fit_model=fit_model)

    def fit(self, x: pd.DataFrame, y: Optional[pd.DataFrame], weights: Optional[pd.Series] = None, fit_preprocessors=True, fit_model=True):
        """
        Fits the model using the given data

        :param x: a data frame containing input data
        :param y: a data frame containing output data; may be None if the underlying model does not actually require
            fitting, e.g. in the case of a rule-based models, but fitting is still necessary for preprocessors
        :param weights: an optional series (with the same index as `x` and `y`) containing data point weights.
            Added in v1.2.0.
        :param fit_preprocessors: whether the model's preprocessors (feature generators and data frame transformers) shall be fitted
        :param fit_model: whether the model itself shall be fitted
        """
        self._trainingContext = TrainingContext(x, y)
        try:
            log.info(f"Fitting {self.__class__.__name__} instance")
            sw = StopWatch()
            self._predictedVariableNames = list(y.columns)
            if not self._underlying_model_requires_fitting():
                if fit_preprocessors:
                    self._fit_preprocessors(x, y=y)
                self._modelInputVariableNames = None  # not known for rule-based models because the fitting process is optimised
            else:
                if y is None:
                    raise Exception(f"The underlying model requires a data frame for fitting but Y=None was passed")
                if len(x) != len(y):
                    raise ValueError(f"Length of input ({len(x)}) does not match length of output ({len(y)})")
                y = self._compute_model_outputs(y)
                x = self._compute_model_inputs(x, y=y, fit=fit_preprocessors)
                if len(x) != len(y):
                    log.debug(f"Input computation changed number of data points ({len(self._trainingContext.original_input)} -> {len(x)})")
                    y = y.loc[x.index]
                    if len(x) != len(y):
                        raise ValueError("Could not recover matching outputs for changed inputs. Only input filtering is admissible, "
                            "indices of input & ouput data frames must match.")
                self._modelInputVariableNames = list(x.columns)
                if fit_model:
                    inputs_with_types = ', '.join([n + '/' + x[n].dtype.name for n in self._modelInputVariableNames])
                    log.debug(f"Fitting with outputs[{len(y.columns)}]={list(y.columns)}, "
                             f"inputs[{len(self._modelInputVariableNames)}]=[{inputs_with_types}]; N={len(x)} data points")
                    self._fit(x, y, **kwarg_if_not_none("weights", weights))
                    self._isFitted = True
                else:
                    log.info("Fitting of underlying model skipped")
            log.info(f"Fitting completed in {sw.get_elapsed_time_secs():.2f} seconds: {self}")
        finally:
            self._trainingContext = None

    def is_being_fitted(self) -> bool:
        """
        :return: True if the model is currently in the process of being fitted, False otherwise
        """
        return self._trainingContext is not None

    @abstractmethod
    def _fit(self, x: pd.DataFrame, y: pd.DataFrame, weights: Optional[pd.DataFrame] = None):
        pass

    def _warn_sample_weights_unsupported(self, is_weighting_supported: bool, weights: Optional[pd.Series]):
        if weights is not None and not is_weighting_supported:
            log.warning(f"Data point weighting not supported by {self.__class__.__name__}; ignoring weights")

    def get_predicted_variable_names(self):
        """
        :return: the list of variable names that are ultimately output by this model (i.e. the columns of the data frame output
            by :meth:`predict`)
        """
        return self._predictedVariableNames

    def get_model_input_variable_names(self) -> Optional[List[str]]:
        """
        :return: the list of variable names required by the underlying model as input (after feature generation and data frame
            transformation) or None if the model has not been fitted (or is a rule-based model which does not determine the variable names).
        """
        return self._modelInputVariableNames

    @deprecated("Use getFeatureTransformer instead, this method will be removed in a future release")
    def get_input_transformer(self, cls: Type[DataFrameTransformer]):
        """
        Gets the (first) feature transformer of the given type (if any) within this models feature transformer chain

        :param cls: the type of transformer to look for
        :return: the first matching feature transformer or None
        """
        for it in self._featureTransformerChain.dataFrameTransformers:
            if isinstance(it, cls):
                return it
        return None

    def get_feature_transformer(self, cls: Type[DataFrameTransformer]):
        """
        Gets the (first) feature transformer of the given type (if any) within this models feature transformer chain

        :param cls: the type of transformer to look for
        :return: the first matching feature transformer or None
        """
        for it in self._featureTransformerChain.dataFrameTransformers:
            if isinstance(it, cls):
                return it
        return None

    def get_raw_input_transformer(self, cls: Type[DataFrameTransformer]):
        """
        Gets the (first) raw input transformer of the given type (if any) within this models raw input transformer chain

        :param cls: the type of transformer to look for
        :return: the first matching raw input transformer or None
        """
        for it in self._rawInputTransformerChain.dataFrameTransformers:
            if isinstance(it, cls):
                return it
        return None

    @deprecated("Use getFeatureTransformerChain instead, this method will be removed in a future release")
    def get_input_transformer_chain(self) -> DataFrameTransformerChain:
        """
        :return: the model's feature transformer chain (which may be empty and contain no actual transformers),
            i.e. the transformers that are applied after feature generation
        """
        return self._featureTransformerChain

    def get_raw_input_transformer_chain(self) -> DataFrameTransformerChain:
        """
        :return: the model's raw input transformer chain (which may be empty and contain no actual transformers),
            i.e. the transformers that are applied before feature generation
        """
        return self._rawInputTransformerChain

    def get_feature_transformer_chain(self) -> DataFrameTransformerChain:
        """
        :return: the model's feature transformer chain (which may be empty and contain no actual transformers),
            i.e. the transformers that are applied after feature generation
        """
        return self._featureTransformerChain

    def set_feature_generator(self, feature_generator: Optional[FeatureGenerator]):
        self.with_feature_generator(feature_generator)

    def get_feature_generator(self) -> Optional[FeatureGenerator]:
        """
        :return: the model's feature generator (if any)
        """
        return self._featureGenerator

    def remove_input_preprocessors(self):
        """
        Removes all input preprocessors (i.e. raw input transformers, feature generators and feature transformers) from the model
        """
        self.with_raw_input_transformers()
        self.with_feature_generator(None)
        self.with_feature_transformers()


class VectorRegressionModel(VectorModel, ABC):
    def __init__(self, check_input_columns=True):
        """
        :param check_input_columns: Whether to check if the input column list (after feature generation)
            during inference coincides with the input column list during fit.
            This should be disabled if feature generation is not performed by the model itself,
            e.g. in ensemble models.
        """
        super().__init__(check_input_columns=check_input_columns)
        self._outputTransformerChain = DataFrameTransformerChain()
        self._modelOutputVariableNames: Optional[list] = None
        self._targetTransformer: Optional[InvertibleDataFrameTransformer] = None

    def _tostring_exclude_exceptions(self) -> List[str]:
        e = super()._tostring_exclude_exceptions()
        if self.TOSTRING_INCLUDE_PREPROCESSORS:
            e += ["_targetTransformer"]
        return e

    def is_regression_model(self) -> bool:
        return True

    def with_output_transformers(self: TVectorRegressionModel,
            *output_transformers: Union[DataFrameTransformer, List[DataFrameTransformer]]) -> TVectorRegressionModel:
        """
        Makes the model use the given output transformers. Call with empty input to remove existing output transformers.
        The transformers are ignored during the fit phase. Not supported for rule-based models.

        **Important**: The output columns names of the last output transformer should be the same
        as the first one's input column names. If this fails to hold, an exception will be raised when :meth:`predict` is called.

        **Note**: Output transformers perform post-processing after the actual predictions have been made. Contrary
        to invertible target transformers, they are not invoked during the fit phase. Therefore, any losses computed there,
        including the losses on validation sets (e.g. for early stopping), will be computed on the non-post-processed data.
        A possible use case for such post-processing is if you know how improve the predictions of your fittable model
        by some heuristics or by hand-crafted rules.

        **How not to use**: Output transformers are not meant to transform the predictions into something with a
        different semantic meaning (e.g. normalized into non-normalized or something like that) - you should consider
        using a targetTransformer for this purpose. Instead, they give the possibility to improve predictions through
        post processing, when this is desired.

        :param output_transformers: DataFrameTransformers for the transformation of outputs
            (after the model has been applied)
        :return: self
        """
        # There is no reason for post-processing in rule-based models
        if not self._underlying_model_requires_fitting():
            raise Exception(f"Output transformers are not supported for model of type {self.__class__.__name__}")
        self._outputTransformerChain = DataFrameTransformerChain(*output_transformers)
        return self

    def with_target_transformer(self: TVectorRegressionModel,
            target_transformer: Optional[InvertibleDataFrameTransformer]) -> TVectorRegressionModel:
        """
        Makes the model use the given target transformers such that the underlying low-level model is trained on the transformed
        targets, but this high-level model still outputs the original (untransformed) values, i.e. the transformation is applied
        to targets during training and the inverse transformation is applied to the underlying model's predictions during inference.
        Hence the requirement of  the transformer being invertible.

        This method is not supported for rule-based models, because they are not trained and therefore the transformation
        would serve no purpose.

        NOTE: All feature generators and data frame transformers - should they make use of outputs - will be fit on the untransformed
        target. The targetTransformer only affects the fitting of the underlying model.

        :param target_transformer: a transformer which transforms the targets (training data outputs) prior to learning the model, such
            that the model learns to predict the transformed outputs
        :return: self
        """
        # Disabled for rule-based models which do not apply fitting and therefore cannot make use of transformed targets
        if not self._underlying_model_requires_fitting():
            raise Exception(f"Target transformers are not supported for model of type {self.__class__.__name__}")
        self._targetTransformer = target_transformer
        return self

    def get_target_transformer(self):
        return self._targetTransformer

    def get_output_transformer_chain(self):
        return self._outputTransformerChain

    def _apply_post_processing(self, y: pd.DataFrame):
        if self._targetTransformer is not None:
            y = self._targetTransformer.apply_inverse(y)
        y = self._outputTransformerChain.apply(y)

        if list(y.columns) != self.get_predicted_variable_names():
            raise Exception(
                f"The model's predicted variable names are not correct. Got "
                f"{list(y.columns)} but expected {self.get_predicted_variable_names()}. "
                f"This kind of error can happen if the model's outputTransformerChain changes a data frame's "
                f"columns (e.g. renames them or changes order). Only output transformer chains that do not change "
                f"columns are permitted in VectorModel. You can fix this by modifying this instance's outputTransformerChain, "
                f"e.g. by calling .withOutputTransformers() with the correct input "
                f"(which can be empty to remove existing output transformers)"
            )
        return y

    def _compute_model_outputs(self, y: pd.DataFrame) -> pd.DataFrame:
        if self._targetTransformer is not None:
            y = self._targetTransformer.fit_apply(y)
        if self.is_being_fitted():
            self._modelOutputVariableNames = list(y.columns)
        return y

    def predict(self, x: pd.DataFrame) -> pd.DataFrame:
        y = super().predict(x)
        return self._apply_post_processing(y)

    def is_fitted(self):
        if not super().is_fitted():
            return False
        if self._targetTransformer is not None and not self._targetTransformer.is_fitted():
            return False
        if not self._outputTransformerChain.is_fitted():
            return False
        return True

    def get_model_output_variable_names(self):
        """
        Gets the list of variable names predicted by the underlying model.
        For the case where at training time the ground truth is transformed by a target transformer
        which changes column names, the names of the variables prior to the transformation will be returned.
        Thus this method always returns the variable names that are actually predicted by the underlying model alone.
        For the variable names that are ultimately output by the entire VectorModel instance when calling predict,
        use getPredictedVariableNames.
        """
        return self._modelOutputVariableNames


def get_predicted_var_name(specified_var_name: Optional[str], predicted_var_names: List[str]):
    if specified_var_name is not None:
        return specified_var_name
    else:
        if len(predicted_var_names) > 1:
            raise ValueError("Must explicitly specify the predicted variable name for a model with multiple output variables "
                f"({predicted_var_names})")
        return predicted_var_names[0]


class VectorClassificationModel(VectorModel, ABC):
    def __init__(self, check_input_columns=True):
        """
        :param check_input_columns: Whether to check if the input column list (after feature generation)
            during inference coincides with the input column list during fit.
            This should be disabled if feature generation is not performed by the model itself,
            e.g. in ensemble models.
        """
        super().__init__(check_input_columns=check_input_columns)
        self._labels = None

    def is_regression_model(self) -> bool:
        return False

    def _fit(self, x: pd.DataFrame, y: pd.DataFrame, weights: Optional[pd.Series] = None):
        if len(y.columns) != 1:
            raise ValueError("Classification requires exactly one output column with class labels")
        self._labels = sorted([label for label in y.iloc[:, 0].unique()])
        self._fit_classifier(x, y, **kwarg_if_not_none("weights", weights))

    def get_class_labels(self) -> List[Any]:
        return self._labels

    @abstractmethod
    def _fit_classifier(self, x: pd.DataFrame, y: pd.DataFrame, weights: Optional[pd.Series] = None):
        pass

    def convert_class_probabilities_to_predictions(self, df: pd.DataFrame):
        """
        Converts from a data frame as returned by predictClassProbabilities to a result as return by predict.

        :param df: the output data frame from predictClassProbabilities
        :return: an output data frame as it would be returned by predict
        """
        labels = self.get_class_labels()
        df_cols = list(df.columns)
        if sorted(df_cols) != labels:
            raise ValueError(f"Expected data frame with columns {labels}, got {df_cols}")
        y_array = df.values
        max_indices = np.argmax(y_array, axis=1)
        result = [df_cols[i] for i in max_indices]
        return pd.DataFrame(result, columns=self.get_predicted_variable_names())

    def predict_class_probabilities(self, x: pd.DataFrame) -> pd.DataFrame:
        """
        :param x: the input data
        :return: a data frame where the list of columns is the list of class labels and the values are probabilities, with the same
            index as the input data frame.
            Raises an exception if the classifier cannot predict probabilities.
        """
        if not self.is_fitted():
            raise Exception(f"Calling predict with unfitted model. "
                            f"This might lead to errors down the line, especially if input/output checks are enabled")
        x = self._compute_model_inputs(x)
        result = self._predict_class_probabilities(x)
        result.index = x.index
        self._check_prediction(result)
        return result

    def _check_prediction(self, prediction_df: pd.DataFrame, max_rows_to_check=5):
        """
        Checks whether the column names are correctly set, sorted and whether the entries correspond to probabilities
        """
        labels = self.get_class_labels()
        if list(prediction_df.columns) != labels:
            raise Exception(f"{self} _predictClassProbabilities returned DataFrame with incorrect columns: "
                            f"expected {labels}, got {prediction_df.columns}")

        df_to_check = prediction_df.iloc[:max_rows_to_check]
        for i, (_, valueSeries) in enumerate(df_to_check.iterrows(), start=1):

            if not all(0 <= valueSeries) or not all(valueSeries <= 1):
                log.warning(f"Probabilities data frame may not be correctly normalised, "
                            f"got probabilities outside the range [0, 1]: checked row {i}/{max_rows_to_check} contains {list(valueSeries)}")

            s = valueSeries.sum()
            if not np.isclose(s, 1, atol=1e-2):
                log.warning(f"Probabilities data frame may not be correctly normalised: "
                    f"checked row {i}/{max_rows_to_check} contains {list(valueSeries)}")

    @abstractmethod
    def _predict_class_probabilities(self, x: pd.DataFrame) -> pd.DataFrame:
        """
        If you are implementing a probabilistic classifier, this method has to return a data frame with probabilities
        (one column per label). The default implementation of _predict will then use the output of
        this method and convert it to predicted labels (via argmax).

        In case you want to predict labels only or have a more efficient implementation of predicting labels than
        using argmax, you may override _predict instead of implementing this method. In the case of a
        non-probabilistic classifier, the implementation of this method should raise an exception.
        """
        raise NotImplementedError(f"{self.__class__.__name__} does not implement _predictClassProbabilities.")

    def _predict(self, x: pd.DataFrame) -> pd.DataFrame:
        predicted_probabilities_df = self._predict_class_probabilities(x)
        return self.convert_class_probabilities_to_predictions(predicted_probabilities_df)


class RuleBasedVectorRegressionModel(VectorRegressionModel, ABC):
    def __init__(self, predicted_variable_names: list):
        """
        :param predicted_variable_names: These are typically known at init time for rule-based models
        """
        super().__init__(check_input_columns=False)
        self._predictedVariableNames = predicted_variable_names
        # guaranteed to be the same as predictedVariableNames since target transformers and output transformers are disallowed
        self._modelOutputVariableNames = predicted_variable_names

    def _underlying_model_requires_fitting(self):
        return False

    def _fit(self, x: pd.DataFrame, y: pd.DataFrame, weights: Optional[pd.Series] = None):
        pass


class RuleBasedVectorClassificationModel(VectorClassificationModel, ABC):
    def __init__(self, labels: list, predicted_variable_name="predictedLabel"):
        """
        :param labels:
        :param predicted_variable_name:
        """
        super().__init__(check_input_columns=False)

        duplicate = get_first_duplicate(labels)
        if duplicate is not None:
            raise Exception(f"Found duplicate label: {duplicate}")
        self._labels = sorted(labels)
        self._predictedVariableNames = [predicted_variable_name]

    def _underlying_model_requires_fitting(self):
        return False

    def _fit(self, x: pd.DataFrame, y: pd.DataFrame, weights: Optional[pd.Series] = None):
        pass

    def _fit_classifier(self, x: pd.DataFrame, y: pd.DataFrame, weights: Optional[pd.Series] = None):
        pass



from enum import Enum
from typing import Union, Callable

import torch
from torch import nn
from torch.nn import functional as F

from sensai.util.pickle import setstate
from ...torch_base import MCDropoutCapableNNModule
from ...torch_enums import ActivationFunction


class LSTNetwork(MCDropoutCapableNNModule):
    """
    Network for (auto-regressive) time-series prediction with long- and short-term dependencies as proposed by G. Lai et al.
    It applies two parallel paths to a time series of size (numInputTimeSlices, inputDimPerTimeSlice):
    
        * Complex path with the following stages:
        
            * Convolutions on the time series input data (CNNs):
              For a CNN with numCnnTimeSlices (= kernel size), it produces an output series of size numInputTimeSlices-numCnnTimeSlices+1.
              If the number of parallel convolutions is numConvolutions, the total output size of this stage is thus
              numConvolutions*(numInputTimeSlices-numCnnTimeSlices+1)
            * Two RNN components which process the CNN output in parallel:
            
                * RNN (GRU)
                  The output dimension of this stage is the hidden state of the GRU after seeing the entire
                  input data from the previous stage, i.e. if has size hidRNN.
                * Skip-RNN (GRU), which processes time series elements that are 'skip' time slices apart.
                  It does this by grouping the input such that 'skip' GRUs are applied in parallel, which all use the same parameters.
                  If the hidden state dimension of each GRU is hidSkip, then the output size of this stage is skip*hidSkip.
                  
            * Dense layer
            
        * Direct regression dense layer (so-called "highway" path) which uses the features of the last hwWindow time slices to
          directly make a prediction
        
    The model ultimately combines the outputs of these two paths via a combination function.
    Many parts of the model are optional and can be completely disabled.
    The model can produce one or more (potentially multi-dimensional) outputs, where each output typically corresponds
    to a time slice for which a prediction is made.

    The model expects as input a tensor of size (batchSize, numInputTimeSlices, inputDimPerTimeSlice).
    As output, the model will produce a tensor of size (batchSize, numOutputTimeSlices, outputDimPerTimeSlice)
    if mode==REGRESSION and a tensor of size (batchSize, outputDimPerTimeSlice=numClasses, numOutputTimeSlices)
    if mode==CLASSIFICATION; the latter shape matches what is required by the multi-dimensional case of loss function
    CrossEntropyLoss, for example, and therefore is suitable for classification use cases.
    For mode==ENCODER, the model will produce a tensor of size (batch_size, hidRNN + skip * hidSkip).
    """

    class Mode(Enum):
        REGRESSION = "regression"
        CLASSIFICATION = "classification"
        ENCODER = "encoder"

    def __init__(self,
            num_input_time_slices: int,
            input_dim_per_time_slice: int,
            num_output_time_slices: int = 1,
            output_dim_per_time_slice: int = 1,
            num_convolutions: int = 100,
            num_cnn_time_slices: int = 6,
            hid_rnn: int = 100,
            skip: int = 0,
            hid_skip: int = 5,
            hw_window: int = 0,
            hw_combine: str = "plus",
            dropout=0.2,
            output_activation: Union[str, ActivationFunction, Callable] = "sigmoid",
            mode: Mode = Mode.REGRESSION):
        """
        :param num_input_time_slices: the number of input time slices
        :param input_dim_per_time_slice: the dimension of the input data per time slice
        :param num_output_time_slices: the number of time slices predicted by the model
        :param output_dim_per_time_slice: the number of dimensions per output time slice. While this is the number of
            target variables per time slice for regression problems, this must be the number of classes for classification problems.
        :param num_cnn_time_slices: the number of time slices considered by each convolution (i.e. it is one of the dimensions of the matrix used for
            convolutions, the other dimension being inputDimPerTimeSlice), a.k.a. "Ck"
        :param num_convolutions: the number of separate convolutions to apply, i.e. the number of independent convolution matrices, a.k.a "hidC";
            if it is 0, then the entire complex processing path is not applied.
        :param hid_rnn: the number of hidden output dimensions for the RNN stage
        :param skip: the number of time slices to skip for the skip-RNN. If it is 0, then the skip-RNN is not used.
        :param hid_skip: the number of output dimensions of each of the skip parallel RNNs
        :param hw_window: the number of time slices from the end of the input time series to consider as input for the highway component.
            If it is 0, the highway component is not used.
        :param hw_combine: {"plus", "product", "bilinear"} the function with which the highway component's output is combined with the complex path's output
        :param dropout: the dropout probability to use during training (dropouts are applied after every major step in the evaluation path)
        :param output_activation: the output activation function
        :param mode: the prediction mode. For `CLASSIFICATION`, the output tensor dimension ordering is adapted to suit loss functions such
            as CrossEntropyLoss. When set to `ENCODER`, will output the latent representation prior to the dense layer in the complex path
            of the network (see class docstring).
        """
        if num_convolutions == 0 and hw_window == 0:
            raise ValueError("No processing paths remain")
        if num_input_time_slices < num_cnn_time_slices or (hw_window != 0 and hw_window < num_input_time_slices):
            raise Exception("Inconsistent numbers of times slices provided")

        super().__init__()
        self.inputDimPerTimeSlice = input_dim_per_time_slice
        self.timeSeriesDimPerTimeSlice = output_dim_per_time_slice
        self.totalOutputDim = self.timeSeriesDimPerTimeSlice * num_output_time_slices
        self.numOutputTimeSlices = num_output_time_slices
        self.window = num_input_time_slices
        self.hidRNN = hid_rnn
        self.numConv = num_convolutions
        self.hidSkip = hid_skip
        self.Ck = num_cnn_time_slices  # the "height" of the CNN filter/kernel; the "width" being inputDimPerTimeSlice
        self.convSeqLength = self.window - self.Ck + 1  # the length of the output sequence produced by the CNN for each kernel matrix
        self.skip = skip
        self.hw = hw_window
        self.pDropout = dropout
        self.mode = mode

        # configure CNN-RNN path
        if self.numConv > 0:
            self.conv1 = nn.Conv2d(1, self.numConv, kernel_size=(self.Ck, self.inputDimPerTimeSlice))  # produce numConv sequences using numConv kernel matrices of size (height=Ck, width=inputDimPerTimeSlice)
            self.GRU1 = nn.GRU(self.numConv, self.hidRNN)
            if self.skip > 0:
                self.skipRnnSeqLength = self.convSeqLength // self.skip  # we divide by skip to obtain the sequence length, because, in order to support skipping via a regrouping of the tensor, the Skip-RNN processes skip entries of the series in parallel to produce skip hidden output vectors
                if self.skipRnnSeqLength == 0:
                    raise Exception("Window size %d is not large enough for skip length %d; would result in Skip-RNN sequence length of 0!" % (self.window, self.skip))
                self.GRUskip = nn.GRU(self.numConv, self.hidSkip)
                self.linear1 = nn.Linear(self.hidRNN + self.skip * self.hidSkip, self.totalOutputDim)
            else:
                self.linear1 = nn.Linear(self.hidRNN, self.totalOutputDim)

        # configure highway component
        if self.hw > 0:
            # direct mapping from all inputs to all outputs
            self.highway = nn.Linear(self.hw * self.inputDimPerTimeSlice, self.totalOutputDim)
            if hw_combine == 'plus':
                self.highwayCombine = self._plus
            elif hw_combine == 'product':
                self.highwayCombine = self._product
            elif hw_combine == 'bilinear':
                self.highwayCombine = nn.Bilinear(self.totalOutputDim, self.totalOutputDim, self.totalOutputDim)
            else:
                raise ValueError("Unknown highway combination function '%s'" % hw_combine)

        self.output = ActivationFunction.torch_function_from_any(output_activation)

    def __setstate__(self, state):
        if "isClassification" in state:
            state["mode"] = self.Mode.CLASSIFICATION if state["isClassification"] else self.Mode.REGRESSION
        setstate(LSTNetwork, self, state, removed_properties=["isClassification"])

    @staticmethod
    def compute_encoder_dim(hid_rnn: int, skip: int, hid_skip: int) -> int:
        return hid_rnn + skip * hid_skip

    def get_encoder_dim(self):
        """
        :return: the vector dimension that is output for the case where mode=ENCODER
        """
        return self.compute_encoder_dim(self.hidRNN, self.skip, self.hidSkip)

    def forward(self, x):
        batch_size = x.size(0)
        # x has size (batch_size, window=numInputTimeSlices, inputDimPerTimeSlice)

        dropout = lambda x: self._dropout(x, p_training=self.pDropout, p_inference=self.pDropout)

        res = None

        if self.numConv > 0:
            # CNN
            # convSeqLength = self.window - self.Ck + 1
            # convolution produces, via numConv kernel matrices of dimension (height=Ck, width=inputDimPerTimeSlice), from an original input sequence of length window, numConv output sequences of length convSeqLength
            c = x.view(batch_size, 1, self.window, self.inputDimPerTimeSlice)  # insert one dim of size 1 (one channel): (batch_size, 1, height=window, width=inputDimPerTimeSlice)
            c = F.relu(self.conv1(c))  # (batch_size, channels=numConv, convSeqLength, 1)
            c = dropout(c)
            c = torch.squeeze(c, 3)  # drops last dimension, i.e. new size (batch_size, numConv, convSeqLength)

            # RNN
            # It processes the numConv sequences of length convSeqLength obtained through convolution and keep the hidden state at the end, which is comprised of hidR entries
            # Specifically, it squashes the numConv sequences of length convSeqLength to a vector of size hidS (by iterating through the sequences and applying the same model in each step, processing all batches in parallel)
            r = c.permute(2, 0, 1).contiguous()  # (convSeqLength, batch_size, numConv)
            self.GRU1.flatten_parameters()
            _, r = self.GRU1(r)  # maps (seq_len=convSeqLength, batch=batch_size, input_size=numConv) -> hidden state (num_layers=1, batch=batch_size, hidden_size=hidR)
            r = torch.squeeze(r, 0)  # (batch_size, hidR)
            r = dropout(r)

            # Skip-RNN
            if self.skip > 0:
                s = c[:, :, -(self.skipRnnSeqLength * self.skip):].contiguous()  # (batch_size, numConv, convSeqLength) -> (batch_size, numConv, skipRnnSeqLength * skip)
                s = s.view(batch_size, self.numConv, self.skipRnnSeqLength, self.skip)  # (batch_size, numConv, skipRnnSeqLength, skip)
                s = s.permute(2, 0, 3, 1).contiguous()  # (skipRnnSeqLength, batch_size, skip, numConv)
                s = s.view(self.skipRnnSeqLength, batch_size * self.skip, self.numConv)  # (skipRnnSeqLength, batch_size * skip, numConv)
                # Why the above view makes sense:
                # skipRnnSeqLength is the sequence length considered for the RNN, i.e. the number of steps that is taken for each sequence.
                # The batch_size*skip elements of the second dimension are all processed in parallel, i.e. there are batch_size*skip RNNs being applied in parallel.
                # By scaling the actual batch size with 'skip', we process 'skip' RNNs of each batch in parallel, such that each RNN consecutively processes entries that are 'skip' steps apart
                self.GRUskip.flatten_parameters()
                _, s = self.GRUskip(s)  # maps (seq_len=skipRnnSeqLength, batch=batch_size * skip, input_size=numConv) -> hidden state (num_layers=1, batch=batch_size * skip, hidden_size=hidS)
                # Because of the way the data is grouped, we obtain not one vector of size hidS but skip vectors of size hidS
                s = s.view(batch_size, self.skip * self.hidSkip)  # regroup by batch -> (batch_size, skip * hidS)
                s = dropout(s)
                r = torch.cat((r, s), 1)  # (batch_size, hidR + skip * hidS)

            if self.mode == self.Mode.ENCODER:
                return r

            res = self.linear1(r)  # (batch_size, totalOutputDim)

        # auto-regressive highway model
        if self.hw > 0:
            resHW = x[:, -self.hw:, :]  # keep only the last hw entries for each input: (batch_size, hw, inputDimPerTimeSlice)
            resHW = resHW.view(-1, self.hw * self.inputDimPerTimeSlice)  # (batch_size, hw * inputDimPerTimeSlice)
            resHW = self.highway(resHW)  # (batch_size, totalOutputDim)
            if res is None:
                res = resHW
            else:
                res = self.highwayCombine(res, resHW)  # (batch_size, totalOutputDim)

        if self.output:
            res = self.output(res)

        res = res.view(batch_size, self.numOutputTimeSlices, self.timeSeriesDimPerTimeSlice)
        if self.mode == self.Mode.CLASSIFICATION:
            res = res.permute(0, 2, 1)
        return res

    @staticmethod
    def _plus(x, y):
        return x + y

    @staticmethod
    def _product(x, y):
        return x * y



import functools
from abc import ABC

import numpy as np
import pandas as pd
from sklearn_quantile import RandomForestQuantileRegressor

from .vector_model import VectorRegressionModel, InputOutputData
from .evaluation.eval_stats import RegressionMetric

from .sklearn.sklearn_base import AbstractSkLearnMultipleOneDimVectorRegressionModel
from .util.aggregation import RelativeFrequencyCounter


class RandomForestQuantileRegressorVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel):
    def __init__(self, confidence: float, random_state=42, **kwargs):
        """
        :param q: the default quantile that is used for predictions
        :param kwargs: keyword arguments to pass on to RandomForestQuantileRegressor
        """
        margin = 1-confidence
        q = [0.5, margin/2, 1 - margin/2]
        super().__init__(RandomForestQuantileRegressor, q=q, random_state=random_state, **kwargs)

    def predict_confidence_intervals(self, x: pd.DataFrame, var_name: str = None):
        """
        :param x: the input data
        :param var_name: the predicted variable name; may be None if there is only one predicted variable
        :return: an array of shape [2, N], where the first dimension contains the confidence interval's lower bounds and the second
            its upper bounds
        """
        model = self.get_sklearn_model(var_name)
        model: RandomForestQuantileRegressor
        outputs = self._predict_quantiles(model, self.compute_model_inputs(x))
        return outputs[1:]

    def _predict_quantiles(self, model: RandomForestQuantileRegressor, inputs: pd.DataFrame) -> np.ndarray:
        outputs = model.predict(inputs)
        return outputs

    def _predict_sklearn_single_model(self, model, inputs: pd.DataFrame) -> np.ndarray:
        return self._predict_quantiles(model, inputs)[0]


class QuantileRegressionMetric(RegressionMetric, ABC):
    @staticmethod
    @functools.lru_cache(maxsize=1)  # use cache for efficient reuse of results across different subclasses during evaluation
    def compute_confidence_intervals(model: VectorRegressionModel, io_data: InputOutputData = None) -> np.ndarray:
        if not isinstance(model, RandomForestQuantileRegressorVectorRegressionModel):
            raise ValueError(f"Model must be of type RandomForestQuantileRegressorVectorRegressionModel, got type {type(model)}")
        intervals: np.ndarray = model.predict_confidence_intervals(io_data.inputs)
        return intervals


class QuantileRegressionMetricAccuracyInConfidenceInterval(QuantileRegressionMetric):
    """
    Metric reflecting the accuracy of the confidence interval, i.e. the relative frequency of predictions where the confidence interval
    contains the ground true value
    """
    name = "AccuracyInCI"

    @classmethod
    def compute_value(cls, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None):
        intervals = cls.compute_confidence_intervals(model, io_data)
        rf = RelativeFrequencyCounter()
        for (lower, upper), gt in zip(intervals.transpose(), y_true):
            rf.count(lower <= gt <= upper)
        return rf.get_relative_frequency()


class QuantileRegressionMetricConfidenceIntervalMeanSize(QuantileRegressionMetric):
    """
    Metric for the mean size of the confidence interval
    """
    name = "MeanSizeCI"

    @classmethod
    def compute_value(cls, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None, io_data: InputOutputData = None):
        intervals = cls.compute_confidence_intervals(model, io_data)
        values = []
        for lower, upper in intervals.transpose():
            values.append(upper-lower)
        return np.mean(values)


class QuantileRegressionMetricConfidenceIntervalMedianSize(QuantileRegressionMetric):
    """
    Metric for the median size of the confidence interval
    """
    name = "MedianSizeCI"

    @classmethod
    def compute_value(cls, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None, io_data: InputOutputData = None):
        intervals = cls.compute_confidence_intervals(model, io_data)
        values = []
        for lower, upper in intervals.transpose():
            values.append(upper-lower)
        return np.median(values)


class QuantileRegressionMetricRelFreqMaxSizeConfidenceInterval(QuantileRegressionMetric):
    """
    Relative frequency of confidence interval having the given maximum size
    """
    def __init__(self, max_size: float):
        super().__init__(f"RelFreqMaxSizeCI[{max_size}]")
        self.max_size = max_size

    def compute_value(self, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None, io_data: InputOutputData = None):
        intervals = self.compute_confidence_intervals(model, io_data)
        counter = RelativeFrequencyCounter()
        for lower, upper in intervals.transpose():
            size = upper-lower
            counter.count(size <= self.max_size)
        return counter.get_relative_frequency()